# -*- coding: utf-8 -*-
"""
Convert FESOM hydrographic output to array at each column point incl lon, lat.

For the coupling of the FESOM ocean model, the original FESOM output of the 
hydrographic fields temp and salt are modified. First values are computed for
all grid point location from the surface layer down to the deepest depth. In
addition for each location the corresponding longitude and latitude information
are provided.

Dependence:
It depends on the pyfesom code (https://github.com/FESOM/pyfesom)


(C) Christian Rodehacke, AWI, 2017-07-12

Additional code for command line parsing:
    Paul Gierz, AWI, May 7, 2018
"""

#
# Command Line Parser
#
import argparse
def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--FESOM_PATH", nargs="*", required=True)
    parser.add_argument("--FESOM_VARIABLE", nargs="+")
    parser.add_argument("--FESOM_MESH", nargs="+")
    parser.add_argument("--FESOM_YEARS", nargs="+")
    parser.add_argument("--FESOM_OUTPUT", nargs="*", default='fesom_output.nc4')
    parser.add_argument("--FESOM_MESH_ALPHA", type=int, default=50)
    parser.add_argument("--FESOM_MESH_BETA",  type=int, default=15)
    parser.add_argument("--FESOM_MESH_GAMMA", type=int, default=-90)
    return parser.parse_args()

args = parse_arguments()

FESOMDATA_OUTPUT = 'fesom_output.nc4'
NAN_REPLACE = 9.96921e36

# ----------------------------------------------------------------
#
# Import external libaries ...
#

import sys
sys.path.append("./")
sys.path.append("./pyfesom")

import pyfesom2 as pf
import numpy as np
from netCDF4 import Dataset, MFDataset
import time as time
import xarray as xr
import os

# ----------------------------------------------------------------
#
# Some information
#

print('* Start at '+time.ctime(time.time()))

# ----------------------------------------------------------------
#
# Read input data
#
# mesh
print('* Read the mesh MESHPATH=' + " ".join(args.FESOM_MESH))

euler_angle = [args.FESOM_MESH_ALPHA, args.FESOM_MESH_BETA, args.FESOM_MESH_GAMMA]

mesh = pf.load_mesh(*args.FESOM_MESH, abg=euler_angle, usepickle=False)
print("*** mesh.zlev = ", mesh.zlev)

#
# Oceanographic data
#
#print('* Read the data file args.FESOM_FILE=' + " ".join(args.FESOM_FILE))
#FID = MFDataset(args.FESOM_PATH + "".join("/temp.fesom.*.01.nc"))
FID = xr.open_dataset(str(args.FESOM_PATH[0]) + str(args.FESOM_VARIABLE[0]) + ".fesom." + str(args.FESOM_YEARS[0]) + ".nc")

# ----------------------------------------------------------------
#
# Initialize size and Preinitialize fields
#
# mesh.n2d = number of 2d nodes, mesh.e2d = number of 2d elements
no_hori_elemnt = mesh.n2d
#no_zlevels = mesh.nlev - 1
no_zlevels = np.size(mesh.zlev)
#no_zlevels = np.size(mesh_diag.zbar)

timedim = 'time' if 'time' in FID.dims else 'T'
no_timesteps = FID.dims[timedim]
print("no_timesteps: ", no_timesteps)
sizeVert = no_zlevels
sizeHori = no_hori_elemnt

sizeFull = (no_timesteps, no_zlevels, no_hori_elemnt)
print("*** sizeFull = ", sizeFull)

# Fields to write: allocate memory upfront
time_out = np.zeros(no_timesteps, dtype=np.float64)
depth_out = np.zeros(sizeVert, dtype=np.float64)
latitude_out = np.zeros(sizeHori, dtype=np.float64)
longitude_out = np.zeros(sizeHori, dtype=np.float64)
TempFields_out = np.zeros(sizeFull, dtype=np.float64)

# ----------------------------------------------------------------
#
# Loop through all depth levels
#
print('* Loop for each time step through all depth levels')

#
# Depth axis
#
ilevel = -1
#for depth in mesh_diag.zbar.squeeze().values:
for depth in mesh.zlev:
    # Some information
    idepth = int(depth)
    #
    # Prepare data for final netcdf output
    #
    ilevel = ilevel + 1
    print('*   ilevel='+str(ilevel)+' depth='+str(idepth)+\
    ' ('+str(depth)+')')
    depth_out[ilevel] = depth

#
# Time axis
#
time_out = FID.variables['time'][0:no_timesteps]

#
# Full hydrographic fields
#

#
# Check shape of input file to process old (2D array) and new (3D array) FESOM files
#
for itime in np.arange(0, no_timesteps, 1, dtype=np.int32):
    ##time_out[itime] = FID.variables['time'][itime]
    time_read = time_out[itime]
    # Some information
    print('*   TIME('+str(itime)+') = '+str(time_read))

    ilevel = -1
    #for depth in mesh_diag.zbar.squeeze().values:
    for depth in mesh.zlev[:-1]:
        # Some information
        idepth = int(depth)
        print('*   depth='+str(idepth)+' ('+str(depth)+\
        ')  ++ time('+str(itime)+') = '+str(time_read))

        #
        # Prepare data for final netcdf output
        #
        ilevel = ilevel + 1

        flag_verbose=False

        if args.FESOM_YEARS[0] == args.FESOM_YEARS[1]:
            level_data = \
            pf.get_data(args.FESOM_PATH[0], args.FESOM_VARIABLE[0], int(args.FESOM_YEARS[0]), mesh, depth=idepth, how="mean") #, flag_verbose)
        else:
            level_data = \
            pf.get_data(args.FESOM_PATH[0], args.FESOM_VARIABLE[0], [int(args.FESOM_YEARS[0]), int(args.FESOM_YEARS[1])], mesh, depth=idepth, how="mean") #, flag_verbose)
        level_data[np.where(np.isnan(level_data))] = NAN_REPLACE
        TempFields_out[itime, ilevel, :] = level_data


# ----------------------------------------------------------------
#
# Output file
#
print('*   Create file "'+args.FESOM_OUTPUT[0]+'"')
#OK: OCE_OUT = Dataset("".join(args.FESOM_OUTPUT), "w", format="NETCDF3_64BIT_OFFSET")
OCE_OUT = Dataset(args.FESOM_OUTPUT[0], "w", format="NETCDF4") #format="NETCDF3_64BIT_OFFSET")

print('*     Create dimensions and define variables')
#
# Create dimensions
#
OCE_OUT.createDimension("time", None)
OCE_OUT.createDimension("level", sizeVert)
OCE_OUT.createDimension("horizontal", no_hori_elemnt)
#
# Create global attributes
#
OCE_OUT.type = "FESOM_ocean_data"
OCE_OUT.title = "FESOM hydrographic ocean data"
OCE_OUT.model_domain = "Generic"
#OCE_OUT.data_input_file = args.FESOM_FILE
OCE_OUT.insitution = 'Alfred Wegener Institute for Polar and Marine Research (AWI)'
OCE_OUT.address = 'Handelshafen, Bremerhaven, Germany'
OCE_OUT.department = "Climate"
OCE_OUT.contact = 'Christian Rodehacke'
OCE_OUT.web = 'http://www.awi.de'
OCE_OUT.acknowledgment = 'ZUWEISS, BMWF, Germany'
OCE_OUT.history = "Created "+time.ctime(time.time())

#
# Create variables
#
# -- axis-variables
time_var = OCE_OUT.createVariable("time", "f4", ("time",))
level_var = OCE_OUT.createVariable("level", "i4", ("level",))
hori_var = OCE_OUT.createVariable("horizontal", "i4", ("horizontal",))

# -- common variables
depth_var = OCE_OUT.createVariable("depth", "f4", ("level",))
lon_var = OCE_OUT.createVariable("longitude", "f4", ("horizontal",))
lat_var = OCE_OUT.createVariable("latitude", "f4", ("horizontal",))

temp_var = OCE_OUT.createVariable(args.FESOM_VARIABLE[0], "f4", \
    ("time", "level", "horizontal",), fill_value=NAN_REPLACE)
#
# Attributes
#
time_var.long_name = "time"
time_var.axis = "T"
time_var.units = "" #FID.variables['time'].units #"hours since 0001-0101 0:0:0.0"
if hasattr(FID.variables['time'], 'calendar'):
    time_var.calendar = FID.variables['time'].calendar #"standard"

level_var.long_name = "model_level"
level_var.description = "Level number"
level_var.axis = "Z"

hori_var.long_name = "horizontal_grid_id"
hori_var.description = "Number of horizontal grid element"
hori_var.axis = "X"

depth_var.long_name = "depth"
depth_var.units = "m"
depth_var.description = "depth below sea level"
depth_var.positive = 'down'

lon_var.long_name = "longitude"
lon_var.units = "degrees east"
lat_var.long_name = "latitude"
lat_var.units = "degrees north"

temp_var.long_name = "temperature"
temp_var.units = "degree celcius" #FID.variables[args.FESOM_VARIABLE[0]].units
temp_var.coordinates = "longitude latitude"
temp_var.description = "" #FID.variables[args.FESOM_VARIABLE[0]].description
#temp_var.missing_value = NAN_REPLACE

#
# Write the data fields
#
print('*     Write the small fields')
# -- axis variables
time_var[:] = time_out
level_var[:] = np.arange(0, sizeVert, dtype=np.int32)
hori_var[:] = np.arange(0, sizeHori, dtype=np.int32)

# -- common variables
depth_var[:] = mesh.zlev
#depth_var[:] = mesh_diag.zbar
lon_var[:] = mesh.x2
lat_var[:] = mesh.y2

print('*     Write the larger fields')
temp_var[:] = TempFields_out

#
# close the files
#
OCE_OUT.close()

#
# Bye bye
#
print('End   at '+time.ctime(time.time()))
print("    ... Bye")



import pandas as pd
import io
import datetime
import numpy as np


def logfile_stats(logfile_to_read):

    with open(logfile_to_read, "r") as f:
        table = f.read()

    df = pd.read_table(
        io.StringIO(table),
        sep=" : | -",
        engine="python",
        names=["Date", "Stuff", "State"],
    )
    # Remove commented rows
    df = df[~df.Stuff.str.startswith("#")]
    # Clean up Date
    df["Date"] = pd.to_datetime(df.Date, errors="coerce")
    # Clean up Stuff
    ## Split columns
    df[["Jobtype", "Job Number", "Job Date", "Job PID"]] = df.Stuff.str.split(
        " ", expand=True
    )
    del df["Stuff"]
    ## Fix types
    df["Jobtype"] = df.Jobtype.str.strip()
    df["Job Number"] = df["Job Number"].astype(int)
    df["Job Date"] = pd.to_datetime(df["Job Date"])
    df["Job PID"] = df["Job PID"].astype(int)
    # Clean up State
    df["State"] = df.State.str.strip()

    while "done" not in df.iloc[-1]["State"]:
        df = df.head(-1)

    gb = df.groupby(["Job Number", "Jobtype"])

    compute_submit_to_start = []
    compute_start_to_end = []
    tidy_start_to_end = []
    queue_time_list = []
    for (jobnumber, jobtype), subdf in gb:
        if jobnumber > 1:
            prev_subdf = gb.get_group((jobnumber - 1, jobtype))
            prev_compute = gb.get_group((jobnumber - 1, "compute"))
            prev_tidy = gb.get_group((jobnumber - 1, "tidy_and_resubmit"))
        if "compute" in jobtype:
            submit = subdf.Date[subdf.State == "submitted"].values[0].astimezone(None)
            end = subdf.Date[subdf.State == "done"].values[0].astimezone(None)
            start = subdf.Date[subdf.State == "start"].values[0].astimezone(None)
            compute_submit_to_start.append((jobnumber, start - submit))
            compute_start_to_end.append((jobnumber, end - start))
            if jobnumber > 1:
                queue_time = start - prev_compute.Date[
                    prev_compute.State == "done"
                ].values[0].astimezone(None)
            else:
                queue_time = None
            queue_time_list.append((jobnumber, queue_time))
        if "tidy" in jobtype:
            end = subdf.Date[subdf.State == "done"].values[0].astimezone(None)
            start = subdf.Date[subdf.State == "start"].values[0].astimezone(None)
            tidy_start_to_end.append((jobnumber, end - start))

    queue_time_df = (
        pd.DataFrame(queue_time_list, columns=["Job Number", "Queue Time"])
        .set_index("Job Number")
        .transpose()
    )

    compute_start_to_end_df = (
        pd.DataFrame(compute_start_to_end, columns=["Job Number", "Compute"])
        .set_index("Job Number")
        .transpose()
    )
    compute_submit_to_start_df = (
        pd.DataFrame(compute_submit_to_start, columns=["Job Number", "Submit"])
        .set_index("Job Number")
        .transpose()
    )
    tidy_df = (
        pd.DataFrame(tidy_start_to_end, columns=["Job Number", "Tidy"])
        .set_index("Job Number")
        .transpose()
    )
    df = pd.concat(
        [queue_time_df, compute_submit_to_start_df, compute_start_to_end_df, tidy_df]
    )
    df = df.transpose()
    print(df)
    print(df.mean())

    one_day = datetime.timedelta(1)
    throughput = one_day / df["Compute"][1:].mean()
    print(
        f"Theoretical Throughput assuming no queueing time: {np.round(throughput, 2)} runs per day"
    )
    return df



import collections.abc
import os
import shutil
import subprocess
import yaml

from loguru import logger


def yprint(pdict):
    """
    Prints variables of a dictionary as a ``yaml.dump``, for pretty printing. Uses the
    ``logger`` from ``loguru`.

    Parameters
    ----------
    pdict : dict
        Dictionary to print
    """
    logger.info(yaml.dump(pdict, default_flow_style=False))


def create_env_loader(tag="!ENV", loader=yaml.SafeLoader):
    """
    Environment loader for ``yaml.load`` to ignore ``!ENV`` variables in the
    ``esm_tests`` methods.

    Returns
    -------
    loader : yaml.loader
    """

    def constructor_env_variables(loader, node):
        return ""

    loader.add_constructor(tag, constructor_env_variables)
    return loader


def sh(inp_str, env_vars=[]):
    """
    Runs a ``shell`` command specified in ``inp_str``, exporting the environment vars
    in ``env_vars``.

    Parameters
    ----------
    inp_str : str
        String containing the ``shell`` command.
    env_vars : list
        List of the environment variables to be exported.

    Returns
    -------
    out : str
        The ``stdout`` and ``stderr`` of the ``shell`` command.
    """
    ev = ""
    # Add the exporting of the environment vars to the command
    for v in env_vars:
        ev += f"export {v}; "
    inp_str = f"{ev}{inp_str}"
    # Run the ``shell`` command
    p = subprocess.Popen(
        inp_str, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True
    )
    out = p.communicate()[0].decode("utf-8")
    return out


def deep_update(d, u):
    """
    Recursively merge dictionary ``u`` into dictionary ``d``. Values of ``u`` win
    always over those of ``d``.

    Parameters
    ----------
    d : dict
        Dictionary in which the merge happens.
    u : dict
        Dictionary that updates ``d``.

    Returns
    -------
    d : dict
        Updated dictionary.
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v

    return d


def copy_comp_files4check_runs(script, script_info, target_dir):
    """
    Gets files from the ``esm_test/resources/runscripts/<model>/comp_files4check_runs``
    so that, for example, ``namelists`` that are hosted in another repository could be
    placed there, to do check runs successfully without having to download the code.

    Parameters
    ----------
    script : str
        Runscript name.
    script_info : dict
        Dictionary containing the information about the runscript.
    target_dir : str
        Target directory path in which the files need to be copied.
    """
    files4check_dir = (
        f"{os.path.dirname(script_info['path'])}/comp_files4check_runs/{script}"
    )
    if os.path.isdir(files4check_dir):
        source_dir = f"{files4check_dir}/{os.listdir(files4check_dir)[0]}"
        combine_folders(source_dir, target_dir)


def combine_folders(source_dir, target_dir):
    """
    Recursively copies the files of the source directory into the target directory,
    including folders and subfolders.

    Parameters
    ----------
    source_dir : str
        Source directory path.
    target_dir : str
        Target directory path.
    """
    # Check that the target path is not a directory
    if not os.path.isfile(target_dir):
        # Check if the source path is a directory
        if os.path.isdir(source_dir):
            # If the target directory does not exist, creates it
            if not os.path.isdir(target_dir):
                os.mkdir(target_dir)
            # Loop through the items inside the source dir and recursively copy them
            # into the target dir
            for folder in os.listdir(source_dir):
                combine_folders(f"{source_dir}/{folder}", f"{target_dir}/{folder}")
        # Check if the source path is a file, it it is copy the file into the target
        if os.path.isfile(source_dir):
            shutil.copy2(source_dir, target_dir)



# MA: This script is not used anymore, it was made on the first place to test the
# environments when I did the refactorization of esm_environments. esm_tests took
# inspiration from some of the functions here, and I decided to keep it here
# just for the record.

import os
import argparse
import subprocess
import re
import shutil
from loguru import logger
import difflib
import yaml
import time
from esm_runscripts import color_diff


def compile_run_all(dir_name, esm_elements, option):

    # Create directory
    if "stable" in dir_name:
        if os.path.isdir(dir_name):
            logger.debug(f"{dir_name} already exists. Avoiding to overwrite.")
            return
    else:
        if os.path.isdir(dir_name):
            delete = input(
                f"{dir_name} already exists, do you want to delete it and calculate it again (y/n)? "
            )
            if delete is "y":
                shutil.rmtree(dir_name)
            else:
                logger.debug(f"{dir_name} already exists. Avoiding to overwrite.")
                return
    os.mkdir(dir_name)

    # Change directory
    top_dir = os.getcwd()
    os.chdir(dir_name)

    # Count total length for timebar
    niter = 0
    for element, versions in esm_elements.items():
        for version in versions:
            niter += 1

    logger.info(f"{dir_name}:")

    # Loop through the different model/setups versions producing the compilation scripts
    c = 0
    rep = 50
    for element, versions in esm_elements.items():
        for version in versions:
            c += 1

            if option == "compile":
                comp_specific(element, version)
            elif option == "run":
                run_specific(element, version)
            else:
                logger.error(f"Option {option} not valid!")

            run = c / niter
            print(
                "|"
                + "\u2588" * round(rep * run)
                + " " * round(rep * (1 - run))
                + "| "
                + str(round(run * 1000) / 10)
                + "%   "
                + element
                + version
                + " " * 20,
                end="\r",
            )

    os.chdir(top_dir)


def comp_specific(element, version):
    cd_format = re.compile("         cd (.*)")
    if version != "":
        version = "-" + version
    # Evaluate and create folders to cheat esm_master
    out = sh("esm_master comp-" + element + version + " -c")
    folders = []
    for line in out.split("\n"):
        if "cd" in line and "cd .." not in line:
            found_format = cd_format.findall(line)
            if len(found_format) > 0:
                if ";" not in found_format[0] and "/" not in found_format[0]:
                    folders.append(found_format[0])
    if len(folders) == 0:
        logger.warning(f'NOT TESTING {element + version}: "cd" command not found')
        return
    prim_f = folders[0]
    folders.append(element + version)
    folders = [x for x in set(folders)]
    if os.path.isdir(prim_f):
        shutil.move(prim_f, prim_f + "_bckp")
    os.mkdir(prim_f)
    for folder in folders:
        os.mkdir(prim_f + "/" + folder)

    # Run the compilation
    out = sh("esm_master comp-" + element + version)

    # Comp files handling
    lfiles = os.listdir()
    file_counter = 0
    for f in lfiles:
        if f.startswith("comp-"):
            if os.path.isfile(prim_f + "/" + f):
                logger.debug(
                    f"{prim_f + '/' + f} already existed. REMOVED and copied again!"
                )
                os.remove(prim_f + "/" + f)
            shutil.move(f, prim_f)
            file_counter += 1

    if file_counter == 0:
        logger.error(f'NOT TESTING {element + version}: no "comp" files')

    # Move directory to have the correct name
    shutil.move(prim_f, element + version)
    if os.path.isdir(prim_f + "_bckp"):
        shutil.move(prim_f + "_bckp", prim_f)


def run_specific(element, version):

    vname = "-" + version
    if len(version) == 0:
        vname = ""

    # Prepare runscript
    runscript = {
        "general": {
            "setup_name": element,
            "version": version,
            "compute_time": "03:00:00",
            "initial_date": "2000-01-01",
            "final_date": "2002-01-01",
            "base_dir": "/work/ollie/mandresm/esm_yaml_test/",
            "account": None,
        },
        element: {
            "model_dir": "nowhere",
        },
    }
    # Prepare the yaml file
    with open("generic_runscript.yaml", "w") as runscript_file:
        yaml.dump(runscript, runscript_file, default_flow_style=False)

    # Run checks
    this_dir = os.getcwd()
    out = sh(
        f"esm_runscripts generic_runscript.yaml -e env_testing -c -v --open-run -U"
    )
    if os.path.isdir("/work/ollie/mandresm/esm_yaml_test/env_testing/scripts"):
        os.chdir("/work/ollie/mandresm/esm_yaml_test/env_testing/scripts")
    else:
        logger.error(f"{element}{vname} check run was not prepared successfully")
        return
    out = sh("esm_runscripts generic_runscript.yaml -e env_testing -c -v --open-run -U")
    os.chdir(this_dir)

    # File copying and cleanup
    os.mkdir(f"{element}{vname}")
    test_path = "/work/ollie/mandresm/esm_yaml_test/env_testing"
    run_path = (
        test_path
        + "/run_20000101-20001231/scripts/env_testing_compute_20000101-20001231.run"
    )
    if not os.path.isfile(run_path):
        logger.error(f"{element}{vname} run file was not generated properly")
    else:
        shutil.copy(run_path, f"{element}{vname}/run-{element}{vname}")
    shutil.rmtree(test_path)


def sh(inp_str):
    p = subprocess.Popen(
        inp_str.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out = p.communicate()[0].decode("utf-8")
    return out


def extract_esm_elements():
    # Extract model/setup names and versions
    out = sh("esm_master")
    out_split_comp = out.split("components:")
    comp_str = out_split_comp[-1]
    setup_str = out_split_comp[0].split("setups:")[-1]
    all_esm = comp_str.split("\n") + setup_str.split("\n")
    esm_elements = {}
    name_format = re.compile("    (.*?):")
    version_format = re.compile("       (.*?):")
    for line in all_esm:
        if len(line) > 4:
            if line[3] == " " and line[4] != " ":
                el_name = name_format.findall(line)[0]
                after_col = line.replace(" ", "")
                if len(after_col) > 0:
                    after_col = after_col.split(":")[1:]
                    if len(after_col) > 1:
                        esm_elements[el_name] = [after_col[0]]
                    elif len(after_col[0]) > 0:
                        esm_elements[el_name] = [""]
                    else:
                        esm_elements[el_name] = []
                else:
                    esm_elements[el_name] = []
            elif len(line) > 7:
                if line[6] == " " and line[7] != " ":
                    version = version_format.findall(line)[0]
                    esm_elements[el_name].append(version)
    return esm_elements


def compare_scripts(stable, test):
    # List all stable folders
    folders = os.listdir(stable)

    # Loop through the elements
    counter = 0
    total = len(folders)
    for element in folders:
        sta = stable + "/" + element
        tes = test + "/" + element
        if not os.path.isdir(tes):
            logger.error(f"{tes} does not exist!")
            continue
        files = os.listdir(sta)
        comps = []
        for f in files:
            if "comp-" in f or "run-" in f:
                comps.append(f)

        logger.info(
            "\n" + len(element) * "=" + f"\n{element}\n" + len(element) * "=" + "\n"
        )
        for comp in comps:
            logger.info(f"\n{comp}\n" + len(comp) * "-" + "\n")
            comp_s = sta + "/" + comp
            comp_t = tes + "/" + comp
            if not os.path.isfile(comp_t):
                logger.error(f"{comp_t} does not exist in {test}!")
                continue
            print_diff(comp_s, comp_t, comp)
        counter += 1
        cont = input(f"Press enter to continue... {str(round(counter/total*1000)/10)}%")


def print_diff(sscript, tscript, name):
    script_s = open(sscript).readlines()
    script_t = open(tscript).readlines()

    diffobj = difflib.SequenceMatcher(a=script_s, b=script_t)
    if diffobj.ratio() == 1:
        logger.info(f"{name} files are identical")
    else:
        differences = ""
        # Find differences
        for line in color_diff(difflib.unified_diff(script_s, script_t)):
            differences += line

        logger.info("Differences:\n" + differences + "\n")


def change_version(version, version_yaml):
    with open(version_yaml, "r") as yaml_file:
        yaml_load = yaml.load(yaml_file, Loader=yaml.FullLoader)
    if not version in yaml_load:
        logger.debug(f"{version} is not in {version_yaml}!")
        sys.exit(-1)

    current_dir = os.getcwd()
    for package, branch in yaml_load[version].items():
        logger.info(f"Installing {branch} for {package}...")
        os.chdir(f"{os.path.expanduser('~')}/{package}")
        sh(f"git checkout {branch}")
        sh("git pull")

    os.chdir(current_dir)


def change_computer(option, version_yaml):
    with open(version_yaml, "r") as yaml_file:
        yaml_load = yaml.load(yaml_file, Loader=yaml.FullLoader)
    computer = yaml_load.get("computer", False)
    if not computer:
        return

    all_machines_path = (
        f"{os.path.expanduser('~')}/esm_tools/configs/machines/all_machines.yaml"
    )

    if option == "change":
        with open(all_machines_path, "r") as machines_file:
            all_machines = yaml.load(machines_file, Loader=yaml.FullLoader)
        this_computer = {computer: all_machines["ollie"]}
        with open(all_machines_path, "w") as machines_file:
            yaml.dump(this_computer, machines_file, default_flow_style=False)
    elif option == "undo":
        this_dir = os.getcwd()
        os.chdir("/".join(all_machines_path.split("/")[:-1]))
        sh("git checkout all_machines.yaml")
        os.chdir(this_dir)


# Get arguments
parser = argparse.ArgumentParser(description="Tests compilation and runscripts")
parser.add_argument("option", default=None)
parser.add_argument("version_yaml", default=None)

args = vars(parser.parse_args())

# Option
option = args["option"]
version_yaml = args["version_yaml"]

# Logging
if os.path.isfile(f"esm_test_{option}.log"):
    os.remove(f"esm_test_{option}.log")
logger.add(f"esm_test_{option}.log", format="{time} {level} {message}")

if option == "compile" or option == "run":
    if option == "compile":
        folder_name = "comp"
    else:
        folder_name = "run"
    # Extract model/setup and version info
    esm_elements = extract_esm_elements()

    # Change to stable version
    change_computer("undo", version_yaml)
    change_version("stable", version_yaml)
    change_computer("change", version_yaml)

    # Compile/run
    compile_run_all(f"{folder_name}_stable", esm_elements, option)

    # Change to new version
    change_computer("undo", version_yaml)
    change_version("test", version_yaml)
    change_computer("change", version_yaml)

    # Compile/run
    compile_run_all(f"{folder_name}_test", esm_elements, option)

    # Print differences
    compare_scripts(f"{folder_name}_stable", f"{folder_name}_test")

elif option == "diff":
    # Print differences
    compare_scripts(f"{version_yaml}_stable", f"{version_yaml}_test")

else:
    print(f'"{option}" is not an option for testing')



import sys

known_couplers = ["oasis3mct", "yac"]


class coupler_class:
    def __init__(self, full_config, name):
        self.name = name

        self.process_ordering = full_config[name]["process_ordering"]
        self.coupled_execs = []
        for exe in self.process_ordering:
            self.coupled_execs.append(full_config[exe]["executable"])
        self.runtime = full_config["general"]["runtime"][5]
        self.nb_of_couplings = 0
        if "coupling_target_fields" in full_config[self.name]:
            for restart_file in list(full_config[self.name]["coupling_target_fields"]):
                self.nb_of_couplings += len(
                    list(full_config[self.name]["coupling_target_fields"][restart_file])
                )
        if "coupling_input_fields" in full_config[self.name]:
            for restart_file in list(full_config[self.name]["coupling_input_fields"]):
                self.nb_of_couplings += len(
                    list(full_config[self.name]["coupling_input_fields"])
                )
        if name == "oasis3mct":
            from . import oasis

            # seb-wahl: manual merge from 'oifs' branch as oifs branch contains many whitespace changes
            self.norestart = full_config["oasis3mct"].get("norestart", "F")
            self.coupler = oasis.oasis(
                self.nb_of_couplings,
                self.coupled_execs,
                self.runtime,
                nnorest=self.norestart,
                mct_version=full_config["oasis3mct"].get("mct_version", "2.8"),
                debug_level=full_config["oasis3mct"].get("debug_level", 1),
                lucia=full_config["oasis3mct"].get("use_lucia", False),
            )
        elif name == "yac":
            from . import yac

            self.coupler = yac.yac(
                full_config,
                self.nb_of_couplings,
                self.process_ordering,
                full_config[self.name]["grids"],
                self.runtime,
            )

        else:
            print("Unknown coupler :", name)
            sys.exit(0)

    def prepare(self, full_config, destination_dir):
        self.add_couplings(full_config)
        self.finalize(destination_dir)
        if full_config["general"]["verbose"]:
            self.print_config_files()

        coupler_name = self.name
        if coupler_name == "yac":
            couplingfile = "coupling.xml"
        else:
            couplingfile = "namcouple"
        return couplingfile

    def print_config_files(self):
        self.coupler.print_config_files()

    def add_files(self, full_config):
        self.coupler.next_coupling = 1
        if "coupling_target_fields" in full_config[self.name]:
            for restart_file in list(full_config[self.name]["coupling_target_fields"]):
                for coupling in full_config[self.name]["coupling_target_fields"][
                    restart_file
                ]:
                    coupling = coupling.replace("<--", "%").replace("--", "&")
                    leftside, rest = coupling.split("%")
                    leftside = leftside.strip()
                    interpolation, rightside = rest.split("&")
                    rightside = rightside.strip()
                    interpolation = interpolation.strip()
                    if ":" in leftside:
                        lefts = leftside.split(":")
                    else:
                        lefts = [leftside]

                    if ":" in rightside:
                        rights = rightside.split(":")
                    else:
                        rights = [rightside]

                    if not len(lefts) == len(rights):
                        print("Left and right side of coupling don't match: ", coupling)
                        sys.exit(0)

                    left_grid = lgrid_info = None
                    right_grid = rgrid_info = None

                    for left, right in zip(lefts, rights):
                        found_left = found_right = False
                        for model in list(full_config):
                            if "coupling_fields" in full_config[model]:
                                if left in full_config[model]["coupling_fields"]:
                                    found_left = True
                                    leftmodel = full_config[model]["executable"]
                                    if not left_grid:
                                        left_grid = full_config[model][
                                            "coupling_fields"
                                        ][left]["grid"]
                                        lgrid_info = full_config[model]["grids"][
                                            left_grid
                                        ]
                                    else:
                                        if (
                                            not left_grid
                                            == full_config[model]["coupling_fields"][
                                                left
                                            ]["grid"]
                                        ):
                                            print(
                                                "All fields coupled together need to exist on same grid"
                                            )
                                            sys.exit(0)
                                if right in full_config[model]["coupling_fields"]:
                                    found_right = True
                                    rightmodel = full_config[model]["executable"]
                                    if not right_grid:
                                        right_grid = full_config[model][
                                            "coupling_fields"
                                        ][right]["grid"]
                                        rgrid_info = full_config[model]["grids"][
                                            right_grid
                                        ]
                                    else:
                                        if (
                                            not right_grid
                                            == full_config[model]["coupling_fields"][
                                                right
                                            ]["grid"]
                                        ):
                                            print(
                                                "All fields coupled together need to exist on same grid"
                                            )
                                            sys.exit(0)
                                if found_right and found_left:
                                    break
                        if not found_left:
                            print("Coupling var not found: ", left)
                        if not found_right:
                            print("Coupling var not found: ", right)
                        if not found_left or not found_right:
                            sys.exit(0)

                    direction_info = None
                    if "coupling_directions" in full_config[self.name]:
                        if (
                            right_grid + "->" + left_grid
                            in full_config[self.name]["coupling_directions"]
                        ):
                            direction_info = full_config[self.name][
                                "coupling_directions"
                            ][right_grid + "->" + left_grid]
                    transf_info = None
                    if "coupling_methods" in full_config[self.name]:
                        if interpolation in full_config[self.name]["coupling_methods"]:
                            transf_info = full_config[self.name]["coupling_methods"][
                                interpolation
                            ]

                    self.coupler.add_output_file(
                        lefts, rights, leftmodel, rightmodel, full_config[self.name]
                    )
                self.coupler.add_restart_files(restart_file, full_config)

    def tidy(self, full_config):
        if full_config[self.name]["lresume"] == False:
            self.prepare_restarts(full_config)

    def prepare_restarts(self, full_config):
        if "coupling_target_fields" in full_config[self.name]:
            for restart_file in list(full_config[self.name]["coupling_target_fields"]):
                all_lefts = []
                all_rights = []
                all_leftmodels = []
                all_rightmodels = []
                for coupling in full_config[self.name]["coupling_target_fields"][
                    restart_file
                ]:
                    coupling = coupling.replace("<--", "%").replace("--", "&")
                    leftside, rest = coupling.split("%")
                    leftside = leftside.strip()
                    interpolation, rightside = rest.split("&")
                    rightside = rightside.strip()
                    if ":" in leftside:
                        lefts = leftside.split(":")
                    else:
                        lefts = [leftside]

                    if ":" in rightside:
                        rights = rightside.split(":")
                    else:
                        rights = [rightside]

                    all_lefts += lefts
                    all_rights += rights

                    for left, right in zip(lefts, rights):
                        found_left = found_right = False
                        for model in list(full_config):
                            if "coupling_fields" in full_config[model]:
                                if left in full_config[model]["coupling_fields"]:
                                    found_left = True
                                    leftmodel = model
                                if right in full_config[model]["coupling_fields"]:
                                    found_right = True
                                    rightmodel = model
                                if found_right and found_left:
                                    break

                    # A coupling restart file can contain fields from multiple models,
                    # therefore, we need to concatenate the left and right models
                    # corresponding to each field
                    all_leftmodels += [leftmodel] * len(lefts)
                    all_rightmodels += [rightmodel] * len(rights)
                    # Check that the dimensions are correct
                    dym_issue = False
                    if len(all_lefts) != len(all_leftmodels):
                        print(
                            "Coupling fields and their corresponding models do not"
                            + "have the same dimensions:"
                        )
                        print("all_lefts =", all_lefts)
                        print("all_leftmodels =", all_leftmodels)
                        dym_issue = True
                    if len(all_rights) != len(all_rightmodels):
                        print(
                            "Coupling fields and their corresponding models do not"
                            + "have the same dimensions:"
                        )
                        print("all_rights =", all_rights)
                        print("all_rightmodels =", all_rightmodels)
                        dym_issue = True
                    if dym_issue:
                        sys.exit(0)

                self.coupler.prepare_restarts(
                    restart_file, all_rights, all_rightmodels, full_config
                )
                self.coupler.prepare_restarts(
                    restart_file + "_recv", all_lefts, all_leftmodels, full_config
                )

    def add_couplings(self, full_config):
        self.coupler.next_coupling = 1
        if self.coupler.name == "oasis3mct":
            if "coupling_target_fields" in full_config[self.name]:
                for restart_file in list(
                    full_config[self.name]["coupling_target_fields"]
                ):
                    for coupling in full_config[self.name]["coupling_target_fields"][
                        restart_file
                    ]:
                        coupling = coupling.replace("<--", "%").replace("--", "&")
                        leftside, rest = coupling.split("%")
                        leftside = leftside.strip()
                        interpolation, rightside = rest.split("&")
                        rightside = rightside.strip()
                        interpolation = interpolation.strip()
                        if ":" in leftside:
                            lefts = leftside.split(":")
                        else:
                            lefts = [leftside]

                        if ":" in rightside:
                            rights = rightside.split(":")
                        else:
                            rights = [rightside]

                        if not len(lefts) == len(rights):
                            print(
                                "Left and right side of coupling don't match: ",
                                coupling,
                            )
                            sys.exit(0)

                        left_grid = lgrid_info = None
                        right_grid = rgrid_info = None

                        for left, right in zip(lefts, rights):
                            found_left = found_right = False
                            for model in list(full_config):
                                if "coupling_fields" in full_config[model]:
                                    if left in full_config[model]["coupling_fields"]:
                                        found_left = True
                                        leftmodel = model
                                        if not left_grid:
                                            left_grid = full_config[model][
                                                "coupling_fields"
                                            ][left]["grid"]
                                            lgrid_info = full_config[model]["grids"][
                                                left_grid
                                            ]
                                        else:
                                            if (
                                                not left_grid
                                                == full_config[model][
                                                    "coupling_fields"
                                                ][left]["grid"]
                                            ):
                                                print(
                                                    "All fields coupled together need to exist on same grid"
                                                )
                                                sys.exit(0)
                                    if right in full_config[model]["coupling_fields"]:
                                        found_right = True
                                        rightmodel = model
                                        if not right_grid:
                                            right_grid = full_config[model][
                                                "coupling_fields"
                                            ][right]["grid"]
                                            rgrid_info = full_config[model]["grids"][
                                                right_grid
                                            ]
                                        else:
                                            if (
                                                not right_grid
                                                == full_config[model][
                                                    "coupling_fields"
                                                ][right]["grid"]
                                            ):
                                                print(
                                                    "All fields coupled together need to exist on same grid"
                                                )
                                                sys.exit(0)
                                    if found_right and found_left:
                                        break
                            if not found_left:
                                print("Coupling var not found: ", left)
                            if not found_right:
                                print("Coupling var not found: ", right)
                            if not found_left or not found_right:
                                sys.exit(0)

                        export_mode = full_config[self.name].get("export_mode", "DEFAULT")

                        direction_info = None
                        if "coupling_directions" in full_config[self.name]:
                            if (
                                right_grid + "->" + left_grid
                                in full_config[self.name]["coupling_directions"]
                            ):
                                direction_info = full_config[self.name][
                                    "coupling_directions"
                                ][right_grid + "->" + left_grid]

                                # Use export_mode from coupling_directions if set. Required for NEMO-AGRIF
                                export_mode = direction_info.get("export_mode",export_mode)
                        transf_info = None
                        if "coupling_methods" in full_config[self.name]:
                            if (
                                interpolation
                                in full_config[self.name]["coupling_methods"]
                            ):
                                transf_info = full_config[self.name][
                                    "coupling_methods"
                                ][interpolation]

                        self.coupler.add_coupling(
                            lefts,
                            lgrid_info,
                            rights,
                            rgrid_info,
                            direction_info,
                            transf_info,
                            restart_file,
                            full_config[self.name]["coupling_time_step"],
                            full_config[self.name]["lresume"],
                            export_mode=export_mode,
                        )

            if "coupling_input_fields" in full_config[self.name]:
                for field_name, field_config in full_config[self.name][
                    "coupling_input_fields"
                ].items():
                    self.coupler.add_input_coupling(
                        field_name, field_config["freq"], field_config["field_filepath"]
                    )

        elif self.coupler.name == "yac":
            if "coupling_directions" in full_config[self.name]:
                transient_id = 1
                for direction in list(full_config[self.name]["coupling_directions"]):
                    for coupling_field in full_config[self.name][
                        "coupling_target_fields"
                    ][direction]:
                        self.coupler.add_coupling(
                            coupling_field,
                            transient_id,
                            direction,
                            full_config[self.name],
                        )
                        transient_id += 1

    def finalize(self, destination_dir):
        self.coupler.finalize(destination_dir)



import os
import copy
import sys

import esm_parser
import esm_rcfile

from . import chunky_parts


def init_first_user_config(command_line_config, user_config):

    if not user_config:
        user_config = get_user_config_from_command_line(command_line_config)

    # maybe switch to another runscript, if iterative coupling
    if user_config["general"].get("iterative_coupling", False):
        user_config = chunky_parts.setup_correct_chunk_config(user_config)

        if len(user_config["general"]["original_config"]["general"]["model_queue"]) > 1:
            next_model = user_config["general"]["original_config"]["general"][
                "model_queue"
            ][1]
        else:
            next_model = user_config["general"]["original_config"]["general"][
                "model_queue"
            ][0]

        scriptname = user_config["general"]["original_config"][next_model]["runscript"]
        # command_line_config["scriptname"] = os.path.join(user_config["general"]["started_from"], scriptname)
        new_command_line_config = copy.deepcopy(command_line_config)
        new_command_line_config["scriptname"] = scriptname
        new_command_line_config["runscript_abspath"] = os.path.join(
            os.getcwd(), scriptname
        )

        model_config = get_user_config_from_command_line(new_command_line_config)
        user_config = esm_parser.new_deep_update(user_config, model_config)

    if user_config["general"].get("debug_obj_init", False):
        pdb.set_trace()

    return user_config


def complete_config_from_user_config(user_config):
    config = get_total_config_from_user_config(user_config)

    if "verbose" not in config["general"]:
        config["general"]["verbose"] = False

    config["general"]["reset_calendar_to_last"] = False

    if config["general"].get("inspect"):
        config["general"]["jobtype"] = "inspect"

        if config["general"].get("inspect") not in [
            "workflow",
            "overview",
            "config",
        ]:
            config["general"]["reset_calendar_to_last"] = True

    return config


def save_command_line_config(config, command_line_config):
    if command_line_config:
        config["general"]["command_line_config"] = command_line_config
    else:
        config["general"]["command_line_config"] = {}

    return config


def get_user_config_from_command_line(command_line_config):
    try:
        # use the full absolute path instead of CWD
        user_config = esm_parser.initialize_from_yaml(
            command_line_config["runscript_abspath"]
        )
        if "additional_files" not in user_config["general"]:
            user_config["general"]["additional_files"] = []
    # If sys.exit is triggered through esm_parser.user_error (i.e. from
    # ``check_for_empty_components`` in ``yaml_to_dict.py``) catch the sys.exit.
    except SystemExit as sysexit:
        sys.exit(sysexit)
    except esm_parser.EsmConfigFileError as error:
        raise error
    except:
        # use the full absolute path instead of CWD
        user_config = esm_parser.initialize_from_shell_script(
            command_line_config["runscript_abspath"]
        )

    # NOTE(PG): I really really don't like this. But I also don't want to
    # re-introduce black/white lists
    #
    # User config wins over command line:
    # -----------------------------------
    # Update all **except** for use_venv if it was supplied in the
    # runscript:
    deupdate_use_venv = False
    if "use_venv" in user_config["general"]:
        user_use_venv = user_config["general"]["use_venv"]
        deupdate_use_venv = True
    user_config["general"].update(command_line_config)
    if deupdate_use_venv:
        user_config["general"]["use_venv"] = user_use_venv
    return user_config


def get_total_config_from_user_config(user_config):

    if "version" in user_config["general"]:
        version = str(user_config["general"]["version"])
    else:
        setup_name = user_config["general"]["setup_name"]

        if "version" in user_config[setup_name.replace("_standalone", "")]:
            version = str(user_config[setup_name.replace("_standalone", "")]["version"])
        else:
            version = "DEFAULT"

    config = esm_parser.ConfigSetup(
        user_config["general"]["setup_name"].replace("_standalone", ""),
        version,
        user_config,
    )

    config = add_esm_runscripts_defaults_to_config(config)

    config["computer"]["jobtype"] = config["general"]["jobtype"]
    config["general"]["experiment_dir"] = (
        config["general"]["base_dir"] + "/" + config["general"]["expid"]
    )

    # Check if the 'account' variable is needed and missing
    if config["computer"].get("accounting", False):
        if "account" not in config["general"]:
            esm_parser.user_error(
                "Missing account info",
                f"You cannot run simulations in '{config['computer']['name']}' "
                "without providing an 'account' variable in the 'general' section, whose "
                "value refers to the project where the computing resources are to be "
                "taken from. Please, add the following to your runscript:\n\n"
                "general:\n\taccount: <the_account_to_be_used>",
            )

    return config


def add_esm_runscripts_defaults_to_config(config):
    FUNCTION_PATH = esm_rcfile.EsmToolsDir("FUNCTION_PATH")
    path_to_file = FUNCTION_PATH + "/esm_software/esm_runscripts/defaults.yaml"
    default_config = esm_parser.yaml_file_to_dict(path_to_file)
    config["general"]["defaults.yaml"] = default_config
    config = distribute_per_model_defaults(config)
    return config


def distribute_per_model_defaults(config):
    default_config = config["general"]["defaults.yaml"]
    if "general" in default_config:
        config["general"] = esm_parser.new_deep_update(
            config["general"], default_config["general"]
        )

    if "per_model_defaults" in default_config:
        per_model_defaults = copy.deepcopy(default_config["per_model_defaults"])
        # Remove ``file_movements`` from per_model_defaults as this is resolved in
        # ``filelists.py`` and otherwise, it is not possible to understand there what
        # comes from the defaults, from general or from the model config itself.
        if "file_movements" in per_model_defaults:
            del per_model_defaults["file_movements"]
        # Set defaults per model
        for model in config["general"]["valid_model_names"]:
            config[model] = esm_parser.new_deep_update(
                config[model], per_model_defaults
            )
    return config



import os
import sys
import psutil
import time

from . import helpers
from . import database_actions
from . import logfiles


def run_job(config):
    helpers.evaluate(config, "assemble", "assemble_recipe")
    return config



def prepare_environment(config):
    environment_dict = {
            "ICE_TO_ECHAM": 1,
            "ECHAM_TO_ICE": 1,
            "ECHAM_TO_ISM_multiyear_mean": 1,
            "ISM_TO_ECHAM_update_orography": 1, 
            "ISM_TO_ECHAM_update_glacial_mask": 1, 
            "ISM_TO_ECHAM_update_land_runoff": 1,
            "COUPLE_DIR": config["general"]["experiment_couple_dir"],
            "RES_echam": config["echam"]["resolution"], 
            "EXP_ID": config["general"]["command_line_config"]["expid"],
            "RESTART_DIR_echam": config["echam"]["experiment_restart_out_dir"],
            "DATA_DIR_echam": config["echam"]["experiment_outdata_dir"],
            "INIT_DIR_echam": config["echam"]["experiment_input_dir"],
            "number_of_years_for_forcing": config["model1"]["chunk_size"],
            "CHUNK_START_DATE_echam": config["general"]["chunk_start_date"],
            "CHUNK_END_DATE_echam": config["general"]["chunk_end_date"],
            "END_YEAR_echam": config["general"]["chunk_end_date"].syear,
            "END_MONTH_echam": config["general"]["chunk_end_date"].smonth,
            "END_DAY_echam": config["general"]["chunk_end_date"].sday,
            "FUNCTION_PATH": "/home/ollie/lackerma/esm_tools/runscripts/awicm-pism/coupling",
            "FORCING_DIR_jsbach": config["jsbach"]["thisrun_forcing_dir"],
            "RESTART_DIR_jsbach": config["jsbach"]["experiment_restart_out_dir"],
            }
    
    #if environment_dict["ADD_UNCHANGED_ICE"] == False:
    #    environment_dict["ADD_UNCHANGED_ICE"] = 0
    #elif environment_dict["ADD_UNCHANGED_ICE"] == True:
    #    environment_dict["ADD_UNCHANGED_ICE"] = 1


    print(environment_dict)
    return environment_dict







import os
import shutil
import subprocess
import copy
import pathlib

import esm_rcfile
import six
import yaml
from esm_calendar import Date
from colorama import Fore, Back, Style, init

import esm_tools
import esm_parser

from .batch_system import batch_system
from .filelists import copy_files, log_used_files
from .helpers import end_it_all, evaluate, write_to_log
from .namelists import Namelist
from loguru import logger

from colorama import Fore


def run_job(config):
    evaluate(config, "prepexp", "prepexp_recipe")
    return config


def color_diff(diff):
    for line in diff:
        if line.startswith("+"):
            yield Fore.GREEN + line + Fore.RESET
        elif line.startswith("-"):
            yield Fore.RED + line + Fore.RESET
        elif line.startswith("^"):
            yield Fore.BLUE + line + Fore.RESET
        else:
            yield line


def copy_tools_to_thisrun(config):
    """
    Copies the tools, namelists and runscripts to the experiment directory,
    making sure that they don't overwrite previously existing files unless
    the ``-U`` flag is used.
    Parameters
    ----------
    config : dict
        Dictionary containing the configuration information.
    """
    gconfig = config["general"]

    # Directory where the original runscript is
    fromdir = os.path.realpath(gconfig["started_from"])
    # Directory where the runscript is copied to
    scriptsdir = os.path.realpath(gconfig["experiment_scripts_dir"])
    # basedir + expid
    expdir = os.path.realpath(gconfig["experiment_dir"])

    # Paths inside the experiment directory where esm_tools and namelists
    # are copied to. Those are not functional but a reference to what was
    # the original state when the experiment was firstly started
    tools_dir = scriptsdir + "/esm_tools/functions"
    namelists_dir = scriptsdir + "/esm_tools/namelists"

    if config["general"]["verbose"]:
        print("Started from :", fromdir)
        print("Scripts Dir : ", scriptsdir)

    # Update namelists and esm_tools. These have no effect on the final
    # simulation as only the installed esm_tools with their runscripts
    # are used, unless esm_tools and namelists paths are manually defined
    # by the model or user (not recommended for the esm_tools path)
    if os.path.isdir(tools_dir) and gconfig["update"]:
        shutil.rmtree(tools_dir, ignore_errors=True)
    if os.path.isdir(namelists_dir) and gconfig["update"]:
        shutil.rmtree(namelists_dir, ignore_errors=True)

    # In case there is no esm_tools or namelists in the experiment folder,
    # copy from the default esm_tools path
    if not os.path.isdir(tools_dir):
        print("Copying standard yamls from: ", esm_rcfile.EsmToolsDir("FUNCTION_PATH"))
        esm_tools.copy_config_folder(tools_dir)
    if not os.path.isdir(namelists_dir):
        print(
            "Copying standard namelists from: ",
            esm_rcfile.EsmToolsDir("NAMELIST_PATH"),
        )
        esm_tools.copy_namelist_folder(namelists_dir)

    # check for recursive creation of the file tree. This prevents the risk of
    # creating a run directory tree inside the `scriptsdir`
    # example:
    # fromdir:    /mnt/lustre02/work/project/username/basedir
    # scriptsdir: /mnt/lustre02/work/project/username/basedir/expid/scripts
    # it is assumed that if 5th parent of the scriptsdir is inside the fromdir
    # then probably some sort of infinite recursion is entered. These lines
    # protect such problems
    scriptsdir_deep_parents = list(pathlib.Path(scriptsdir).parents)[5:]
    deep_nesting_found = pathlib.Path(expdir) in scriptsdir_deep_parents
    if deep_nesting_found:
        error_type = "runtime error"
        error_text = (
            f"deep recursion is detected in {__file__}:\n"
            f"- scriptsdir:         {scriptsdir}\n"
            f"- fromdir:            {fromdir}\n"
            f"- experiment dir:     {expdir}"
        )
        # exit right away to prevent further recursion. There might still be
        # running instances of esmr_runscripts and something like
        # `killall esm_runscripts` might be required
        esm_parser.user_error(error_type, error_text)

    # If ``fromdir`` and ``scriptsdir`` are the same, this is already a computing
    # simulation which means we want to use the script in the experiment folder,
    # so no copying is needed
    if (fromdir == scriptsdir) and not gconfig["update"]:
        if config["general"]["verbose"]:
            print("Started from the experiment folder, continuing...")
        return config
    # Not computing but initialisation
    else:
        if not fromdir == scriptsdir:
            if config["general"]["verbose"]:
                print("Not started from experiment folder, restarting...")
        else:
            print("Tools were updated, restarting...")

        # At this point, ``fromdir`` and ``scriptsdir`` are different. Update the
        # runscript if necessary
        update_runscript(
            fromdir, scriptsdir, gconfig["scriptname"], gconfig, "runscript"
        )

        if gconfig.get("iterative_coupling", False):
            index = 1
            while "model" + str(index) in config:
                update_runscript(
                    fromdir,
                    scriptsdir,
                    config["model" + str(index)]["runscript"],
                    gconfig,
                    "runscript",
                )
                index += 1

        # Update the ``additional_files`` if necessary
        for tfile in gconfig["additional_files"]:
            update_runscript(fromdir, scriptsdir, tfile, gconfig, "additional file")

        # remove the update option otherwise it will enter an infinite loop
        original_command = gconfig["original_command"]
        options_to_remove = [" -U ", " --update "]
        for option in options_to_remove:
            original_command = original_command.replace(option, " ")

        # Before resubmitting the esm_runscripts, the path of the runscript
        # needs to be modified. Remove the absolute/relative path
        runscript_absdir, runscript = os.path.split(gconfig["runscript_abspath"])
        original_command_list = original_command.split()
        new_command_list = []
        for command in original_command_list:
            # current command will contain the full path, so replace it with
            # the YAML file only since we are going to execute it from the
            # `scriptsdir` now
            if runscript in command:
                # gconfig['scriptname'] or `runscript` only contains the YAML file name
                command = runscript
            new_command_list.append(command)

        new_command = " ".join(new_command_list)
        restart_command = f"cd {scriptsdir}; esm_runscripts {new_command}"

        # prevent continuous addition of --no-motd
        if not "--no-motd" in restart_command:
            restart_command += " --no-motd "

        if config["general"]["verbose"]:
            print(restart_command)
        os.system(restart_command)

        gconfig["profile"] = False
        end_it_all(config)


def _create_folders(config, filetypes):
    """
    Generates the experiment file tree. Foldres are created for every filetype
    except for "ignore".
    """
    for filetype in filetypes:
        if not filetype == "ignore":
            if not filetype == "work":
                if not os.path.exists(config["experiment_" + filetype + "_dir"]):
                    os.makedirs(config["experiment_" + filetype + "_dir"])
            if not os.path.exists(config["thisrun_" + filetype + "_dir"]):
                os.makedirs(config["thisrun_" + filetype + "_dir"])


def _create_setup_folders(config):
    """
    Creates all folders for an experiment. This is the plugin function called
    during the recipe. See also ``_create_folders`` for actual creation

    This also creates a small marker file at the top of
    the experiment so that the "root" can be found from inside.
    """
    _create_folders(config["general"], config["general"]["all_filetypes"])
    with open(
        config["general"]["experiment_dir"] + "/.top_of_exp_tree", "w"
    ) as top_marker:
        top_marker.write(f"Top of experiment {config['general']['expid']}")
    return config


def _create_component_folders(config):
    for component in config["general"]["valid_model_names"]:
        _create_folders(config[component], config["general"]["all_model_filetypes"])
    return config


def initialize_experiment_logfile(config):
    """
    Initializes the log file for the entire experiment.

    Creates a file ``${BASE_DIR}/${EXPID}/log/${EXPID}_${setup_name}.log``
    to keep track of start/stop times, job id numbers, and so on. Use the
    function ``write_to_log`` to put information in this file afterwards.

    The user can specify ``experiment_log_file`` under the ``general``
    section of the configuration to override the default name. Timestamps
    for each message are given by the section
    ``experiment_log_file_dateformat``, or defaults to ``Tue Mar 17
    09:36:38 2020``, i.e. ``"%c"``. Please use ``stftime`` compatable
    formats, as described here: https://strftime.org

    Parameters
    ----------
    dict :
        The experiment configuration

    Return
    ------
    dict :
        As per convention for the plug-in system; this gives back the
        entire config.

    Attention
    ---------
        Calling this has some filesystem side effects. If the run number in
        the general configuration is set to 1, and a file exists for
        ``general.exp_log_file``; this file is removed; and re-initialized.
    """

    if config["general"]["run_number"] == 1:
        if os.path.isfile(config["general"]["experiment_log_file"]):
            os.remove(config["general"]["experiment_log_file"])

        log_msg = f"# Beginning of Experiment {config['general']['expid']}"
        write_to_log(config, [log_msg], message_sep="")

        write_to_log(
            config,
            [
                str(config["general"]["jobtype"]),
                str(config["general"]["run_number"]),
                str(config["general"]["current_date"]),
                str(config["general"]["jobid"]),
                "- start",
            ],
        )

    # Write trace-log file now that we know where to do that
    if "trace_sink" in dir(logger):
        logfile_path = (
            f"{config['general']['experiment_dir']}/log"
            f"/{config['general']['expid']}_esm_runscripts_"
            f"{config['general']['run_datestamp']}.log"
        )

        logger.trace_sink.def_path(logfile_path)

    return config


def update_runscript(fromdir, scriptsdir, tfile, gconfig, file_type):
    """
    Updates the script ``tfile`` in the directory ``scriptdir`` with the file in
    the directory ``fromdir`` if the update flag (``-U``) is used during the call of
    ``esm_runscripts``. If that flag is not used and the source and target are different
    then raises a user-friendly error recommending to use the ``-U`` flag with the warning
    that the files will be overwritten.
    Parameters
    ----------
    cls : obj
        Compute object.
    fromdir : str
        Path of the source.
    scriptsdir : str
        Path of the target.
    tfile : str
        Name of the script to be updated.
    gconfig : dict
        Dictionary containing the general information about the compute task.
    file_type : str
        String specifying the nature of the file, only necessary for printing information
        and for the error description.
    Exceptions
    ----------
    UserError
        If the target and source are different and the ``-U`` flag is not used when calling
        ``esm_runscripts``, returns an error.
    """

    # if `tfile` contains a full path of the runscript then remove the leading path
    tfile = os.path.basename(tfile)

    # If the target file in ``scriptsdir`` does not exist, then copy the file
    # to the target.
    if not os.path.isfile(scriptsdir + "/" + tfile):
        oldscript = fromdir + "/" + tfile
        print(oldscript)
        shutil.copy2(oldscript, scriptsdir)
    # If the target path exists compare the two scripts
    else:
        import difflib
        import esm_parser

        script_o = open(fromdir + "/" + tfile).readlines()
        script_t = open(scriptsdir + "/" + tfile).readlines()

        diffobj = difflib.SequenceMatcher(a=script_t, b=script_o)
        # If the files are different
        if not diffobj.ratio() == 1:
            # Find differences
            differences = (
                f"{fromdir + '/' + tfile} differs from "
                + f"{scriptsdir + '/' + tfile}:\n"
            )
            for line in color_diff(difflib.unified_diff(script_t, script_o)):
                differences += line

            # If the --update flag is used, notify that the target script will
            # be updated and do update it
            if gconfig["update"]:
                esm_parser.user_note(
                    f"Original {file_type} different from target",
                    differences + "\n" + f"{scriptsdir + '/' + tfile} will be updated!",
                )
                oldscript = fromdir + "/" + tfile
                print(oldscript)
                shutil.copy2(oldscript, scriptsdir)
            # If the --update flag is not called, exit with an error showing the
            # user how to proceed
            else:
                esm_parser.user_note(
                    f"Original {file_type} different from target",
                    differences
                    + "\n"
                    + "Note: You can choose to use -U flag in the esm_runscripts call "
                    + "to automatically update the runscript (WARNING: This "
                    + f"will overwrite your {file_type} in the experiment folder!)\n",
                )
                correct_input = False
                while not correct_input:
                    update_choice = input(
                        f"Do you want that {scriptsdir + '/' + tfile} is "
                        + "updated with the above changes? (y/n): "
                    )
                    if update_choice == "y":
                        correct_input = True
                        oldscript = fromdir + "/" + tfile
                        print(oldscript)
                        shutil.copy2(oldscript, scriptsdir)
                        print(f"{scriptsdir + '/' + tfile} updated!")
                    elif update_choice == "n":
                        correct_input = True
                        esm_parser.user_error(
                            f"Original {file_type} different from target",
                            differences
                            + "\n"
                            + "You can choose to -U flag in the esm_runscripts call "
                            + "to update the runscript without asking (WARNING: This "
                            + f"will overwrite your {file_type} in the experiment folder!)\n\n",
                        )
                    else:
                        print(f"'{update_choice}' is not a valid answer.")


def _copy_preliminary_files_from_experiment_to_thisrun(config):
    # I don't like this one bit. DB
    filelist = [
        (
            "scripts",
            f"{config['general']['expid']}_{config['general']['setup_name']}.date",
            "copy",
        )
    ]

    for filetype, filename, copy_or_link in filelist:
        source = config["general"]["experiment_" + filetype + "_dir"]
        dest = config["general"]["thisrun_" + filetype + "_dir"]
        if copy_or_link == "copy":
            method = shutil.copy2
        elif copy_or_link == "link":
            method = os.symlink
        if os.path.isfile(source + "/" + filename):
            method(source + "/" + filename, dest + "/" + filename)
    this_script = config["general"]["scriptname"]
    shutil.copy2("./" + this_script, config["general"]["thisrun_scripts_dir"])

    for additional_file in config["general"]["additional_files"]:
        shutil.copy2(additional_file, config["general"]["thisrun_scripts_dir"])
    return config



######################################################################################
############################## class "software_package" ##############################
######################################################################################
import sys


def replace_var(var, tag, value):
    if var and tag and value:
        # deniz, TODO: In the future I will implement a better way around
        if isinstance(var, (str, int, float)):
            var = str(var)
            return var.replace("${" + tag + "}", value)
        elif type(var) == list:
            newlist = []
            for entry in var:
                newlist.append(replace_var(entry, tag, value))
            return newlist
        else:
            print(f"ERROR: unresolved variable: {var}")
            sys.exit(1)


class software_package:
    def __init__(
        self, raw, setup_info, vcs, general, no_infos=False
    ):  # model_and_version):
        if type(raw) == str:
            (
                dummy,
                self.kind,
                self.model,
                self.version,
                dummy2,
                self.raw_name,
            ) = setup_info.split_raw_target(raw, setup_info)
        else:  # tupel:
            (self.kind, self.model, self.version) = raw
            self.raw_name = setup_info.assemble_raw_name(
                None, self.kind, self.model, self.version
            )

        self.tag = None

        # deniz: Linux pipe support: eg. curl foo.tar.gz | tar zx
        # !!! if these lines are not here esm_master crashes on xios.
        self.pipe_options = None

        # deniz: I don't like the following lines. This is not good OOP.
        # Constructor should initialize all variables and they should not be
        # defined in an else block
        if not no_infos:
            self.fill_in_infos(setup_info, vcs, general)
        else:
            self.targets = self.subpackages = None

            # kh 11.09.20 support git options like --recursive
            self.repo_type = None
            self.repo = None
            self.branch = None
            self.repo_options = None
            self.bin_type = None
            self.bin_names = [None]
            self.command_list = None
            self.destination = None
            self.clone_destination = None
            self.coupling_changes = None
            # deniz: Linux pipe support: eg. curl foo.tar.gz | tar zx
            self.pipe_options = None

    def fill_in_infos(self, setup_info, vcs, general):

        self.targets = self.get_targets(setup_info, vcs)
        self.subpackages = self.get_subpackages(setup_info, vcs, general)
        self.complete_targets(setup_info)

        # kh 11.09.20 support git options like --recursive
        self.repo_type, self.repo, self.branch, self.repo_options = self.get_repo_info(
            setup_info, vcs
        )
        self.destination = setup_info.get_config_entry(self, "destination")
        self.clone_destination = setup_info.get_config_entry(self, "clone_destination")
        if not self.destination:
            self.destination = self.raw_name

        self.coupling_changes = self.get_coupling_changes(setup_info)
        self.repo = replace_var(self.repo, self.model + ".version", self.version)
        self.branch = replace_var(self.branch, self.model + ".version", self.version)
        self.repo_options = replace_var(
            self.repo_options, self.model + ".version", self.version
        )

        # deniz: Linux pipe support
        self.pipe_options = setup_info.get_config_entry(self, "pipe_options")

        self.bin_type, self.bin_names = self.get_comp_type(setup_info)
        self.command_list = self.get_command_list(setup_info, vcs, general)

    def get_targets(self, setup_info, vcs):
        config = setup_info.config
        targets = []
        for todo in setup_info.known_todos:
            if setup_info.has_target(self, todo, vcs):
                targets.append(todo)
        return targets

    def complete_targets(self, setup_info):
        for todo in setup_info.known_todos:
            for package in self.subpackages:
                if todo in package.targets:
                    if todo not in self.targets:
                        self.targets.append(todo)

    def get_coupling_changes(self, setup_info):
        config = setup_info.config
        changes = []
        if self.kind == "couplings":
            these_changes = setup_info.get_config_entry(self, "coupling_changes")
            if these_changes:
                changes = changes + these_changes
        elif self.kind == "setups":
            couplings = setup_info.get_config_entry(self, "couplings")
            if couplings:
                for coupling in couplings:
                    changes = []
                    if "coupling_changes" in config["couplings"][coupling]:
                        these_changes = config["couplings"][coupling][
                            "coupling_changes"
                        ]
                        if these_changes:
                            changes = changes + these_changes
        return changes

    def get_subpackages(self, setup_info, vcs, general):
        subpackages = []
        config = setup_info.config

        if self.kind == "setups":
            couplings = setup_info.get_config_entry(self, "couplings")
            if couplings:
                for coupling in couplings:
                    newpackage = software_package(coupling, setup_info, vcs, general)
                    subpackages += newpackage.get_subpackages(setup_info, vcs, general)

        if self.kind == "couplings":
            components = setup_info.get_config_entry(self, "components")
            if components:
                for component in components:
                    found = False
                    for package in subpackages:
                        if component == package.raw_name:
                            found = True
                            break
                    if not found:
                        newpackage = software_package(
                            component, setup_info, vcs, general
                        )
                        subpackages += newpackage.get_subpackages(
                            setup_info, vcs, general
                        )
                        subpackages.append(newpackage)

        elif self.kind == "components":
            requirements = setup_info.get_config_entry(self, "requires")
            if requirements:
                for component in requirements:
                    found = False
                    for package in subpackages:
                        if component == package.raw_name:
                            found = True
                            break
                    if not found:
                        newpackage = software_package(
                            component, setup_info, vcs, general
                        )
                        subpackages += newpackage.get_subpackages(
                            setup_info, vcs, general
                        )
                        subpackages.append(newpackage)

        # if self.kind == "couplings":
        #    components = setup_info.get_config_entry(self, "components")
        #    if components:
        #        for component in components:
        #            comp_tupel = software_package(component, setup_info, vcs, general)
        #            if comp_tupel not in subpackages:
        #                subpackages.append(comp_tupel)
        # elif self.kind == "setups":
        #    couplings = setup_info.get_config_entry(self, "couplings")
        #    if couplings:
        #        for coupling in couplings:
        #            for component in config["couplings"][coupling]["components"]:
        #                 found = False
        #                 for package in subpackages:
        #                    if component == package.raw_name:
        #                        found = True
        #                        break
        #                if not found:
        #                    subpackages.append(
        #                        software_package(component, setup_info, vcs, general)
        #                    )
        # elif self.kind == "components":
        #    requirements = setup_info.get_config_entry(self, "requires")
        ##    if requirements:
        #        for requirement in requirements:
        #            found = False
        #            if subpackages:
        #                for package in subpackages:
        #                    if requirement == package.raw_name:
        #                        found = True
        #                        break
        #                if not found:
        #                    subpackages.append(
        #                      software_package(requirement, setup_info, vcs, general)
        #                    )
        #            else:
        #                subpackages.append(
        #                  software_package(requirement, setup_info, vcs, general)
        #                )
        return subpackages

    def get_comp_type(self, setup_info):
        exec_names = setup_info.get_config_entry(self, "install_bins")
        if exec_names:
            if type(exec_names) == str:
                exec_names = [exec_names]
            return "bin", exec_names
        exec_names = setup_info.get_config_entry(self, "install_libs")
        if exec_names:
            if type(exec_names) == str:
                exec_names = [exec_names]
            return "lib", exec_names
        return "bin", []

    def get_repo_info(self, setup_info, vcs):
        repo = branch = repo_type = None
        for check_repo in vcs.known_repos:
            repo = setup_info.get_config_entry(self, check_repo + "-repository")
            if repo:
                repo_type = check_repo
                break
        branch = setup_info.get_config_entry(self, "branch")

        # kh 11.09.20 support git options like --recursive
        repo_options = setup_info.get_config_entry(self, "repo_options")

        return repo_type, repo, branch, repo_options

    def get_command_list(self, setup_info, vcs, general):
        command_list = {}
        for todo in self.targets:
            if todo in vcs.known_todos:
                commands = vcs.assemble_command(self, todo, setup_info, general)
            else:
                commands = setup_info.get_config_entry(self, todo + "_command")
            if commands:
                if type(commands) == str:
                    commands = [commands]
                if not todo == "get":
                    commands.insert(0, "cd " + self.destination)
                    commands.append("cd ..")
            if todo == "get":
                if self.coupling_changes:
                    commands = []
                    for change in self.coupling_changes:
                        commands.append(change)
            command_list.update({todo: commands})
        return command_list

    def output(self):
        print()
        print(self.raw_name)
        print(
            "    Model:", self.model, ", Version:", self.version, ", Kind:", self.kind
        )
        if self.subpackages:
            for package in self.subpackages:
                print("    Subpackage: ", package.raw_name)
        if self.targets:
            print("    Targets: ", self.targets)
        if self.repo_type:
            print(
                "    Repo_type: ",
                self.repo_type,
                ", repo: ",
                self.repo,
                ", branch: ",
                self.branch,
                # kh 11.09.20 support git options like --recursive
                ", repo_options: ",
                self.repo_options,
            )
        if self.bin_type:
            print("    Bin_type: ", self.bin_type, ", bin_names: ", self.bin_names)
        if self.command_list:
            print("    Commands:")
            for todo in self.command_list.keys():
                print("        ", todo, self.command_list[todo])
        if self.coupling_changes:
            print("    Coupling Changes:")
            for todo in self.coupling_changes:
                print("        ", todo)



import random
import numpy as np
import xarray as xr
import pandas as pd
import sys
import math
import os
from tqdm import tqdm
from datetime import datetime
from numpy.linalg import inv

mesh_path = sys.argv[1]
opath = sys.argv[2]
nod2d_file = os.path.join(mesh_path, "nod2d.out")

def _generate_rotation_matrix():
    alphaEuler=50
    betaEuler=15
    gammaEuler=-90
    
    al = np.radians(alphaEuler)
    be = np.radians(betaEuler)
    ga = np.radians(gammaEuler)
   
    rotate_matrix = np.zeros((3,3))
    rotate_matrix[0,0]=np.cos(ga)*np.cos(al)-np.sin(ga)*np.cos(be)*np.sin(al)
    rotate_matrix[0,1]=np.cos(ga)*np.sin(al)+np.sin(ga)*np.cos(be)*np.cos(al)
    rotate_matrix[0,2]=np.sin(ga)*np.sin(be)
    rotate_matrix[1,0]=-np.sin(ga)*np.cos(al)-np.cos(ga)*np.cos(be)*np.sin(al)
    rotate_matrix[1,1]=-np.sin(ga)*np.sin(al)+np.cos(ga)*np.cos(be)*np.cos(al)
    rotate_matrix[1,2]=np.cos(ga)*np.sin(be)
    rotate_matrix[2,0]=np.sin(be)*np.sin(al) 
    rotate_matrix[2,1]=-np.sin(be)*np.cos(al)  
    rotate_matrix[2,2]=np.cos(be)
    
    return rotate_matrix

def _r2g(lon_r, lat_r):
    A = _generate_rotation_matrix()
    A_ = inv(A)

    lon_ = np.radians(lon_r)
    lat_ = np.radians(lat_r)

    v_ = np.zeros((3,1))
    v_[0]=np.cos(lat_)*np.cos(lon_)
    v_[1]=np.cos(lat_)*np.sin(lon_)
    v_[2]=np.sin(lat_) 
    vr = np.dot(A_, v_)

    lon_g = np.degrees(math.atan2(vr[1], vr[0]))
    lat_g = np.degrees(math.asin(vr[2]))
    return  lon_g, lat_g

def _back_rotate():
    r2g_v = np.vectorize(_r2g, excluded="A")

    with open(nod2d_file, 'r') as csvfile:
        nodes = pd.read_csv(csvfile, header=None, sep=r'\s* \s*', skiprows=1, engine='python')
        #nodes.drop(nodes.index[0], inplace=True)
        nodes.columns = ['index', 'lon', 'lat', 'mask']
        [lon_tmp, lat_tmp] = r2g_v(np.array(nodes['lon'].values[:], dtype=float).transpose(), 
                                   np.array(nodes['lat'].values[:], dtype=float).transpose())
        nodes['lon'] = lon_tmp
        nodes['lat'] = lat_tmp
        nodes.to_csv(os.path.join(opath, "nod2d.out"), sep=' ', header=[str(len(nodes)), "", "", ""], index=False)

_back_rotate()



import os
import copy
import esm_parser


class last_minute_changes:
    def __init__(self, config):
        self.modify_config_file = config["general"].get("modify_config_file")

        if self.modify_config_file:
            self.modify_config_file_abspath = os.path.abspath(self.modify_config_file)
            self.modify_config = esm_parser.yaml_file_to_dict(
                self.modify_config_file_abspath
            )

            config["general"]["modify_config"] = copy.deepcopy(self.modify_config)
            config["general"][
                "modify_config_file_abspath"
            ] = self.modify_config_file_abspath

            # kh 27.11.20 "original command" is not available for esm_master (but for esm_runscripts)
            if "original_command" in "general":
                config["general"]["original_command"] = config["general"][
                    "original_command"
                ].replace(self.modify_config_file, self.modify_config_file_abspath)

        else:
            self.modify_config_file_abspath = self.modify_config = None


def apply_last_minute_changes(config):
    config["general"]["modify_config_memo"] = last_minute_changes(config)

    modify_config = config["general"]["modify_config_memo"].modify_config

    if modify_config:
        settings = (
            modify_config.get("build_and_run_modifications", {})
            .get("machine", {})
            .get("chooseable_settings")
        )
        _modify_config_with_settings(config, settings)

        settings = (
            modify_config.get("build_only_modifications", {})
            .get("machine", {})
            .get("environment_settings")
        )
        _modify_config_with_settings(config, settings)

        settings = (
            modify_config.get("run_only_modifications", {})
            .get("machine", {})
            .get("chooseable_settings")
        )
        _modify_config_with_settings(config, settings)

        settings = (
            modify_config.get("run_only_modifications", {})
            .get("batch_system", {})
            .get("direct_settings")
        )
        _modify_config_with_settings(config, settings)

        for chapter in modify_config:
            if chapter not in [
                "build_and_run_modifications",
                "build_only_modifications",
                "run_only_modifications",
            ]:
                esm_parser.dict_merge(config, modify_config)

    return config


def restore_protected_last_minute_changes(config):
    if config["general"]["modify_config_memo"]:
        if config["general"]["modify_config_memo"].modify_config:
            config["general"]["modify_config"] = config["general"][
                "modify_config_memo"
            ].modify_config
        del config["general"]["modify_config_memo"]

    # kh 26.11.20
    if "modify_config_memo" in config["general"]:  # Entry could exist but be False
        del config["general"]["modify_config_memo"]

    return config


def _modify_config_with_settings(config, settings):
    if settings:
        for k, v in settings.items():
            path_to_key = k.split(".")
            entry = path_to_key.pop()
            selected_config = config
            for k2 in path_to_key:
                selected_config = selected_config[k2]
            if type(selected_config) == dict:
                selected_config[entry] = v
            elif type(selected_config) == list:
                selected_config.append(entry + "=" + v)

            else:
                raise ValueError("unexpected container type (neither dict nor list")



#!/usr/bin/env python
"""
Main module for EsmEnvironment.
"""

import copy
import os
import warnings
import re
import sys

import esm_parser
from esm_rcfile import FUNCTION_PATH

######################################################################################
########################### class "environment_infos" ################################
######################################################################################


class EnvironmentInfos:
    """
    The package ``esm_environment`` takes care of generating the environments for the
    different HPCs supported by `ESM-Tools`. This is done through the use of the
    ``EnvironmentInfos`` class inside the different `ESM-Tools` packages.

    For the correct definition of an environment for an HPC a `yaml` file for that
    system needs to be included inside the ``esm_tools`` package inside the
    ``configs/machines/`` folder (e.g. ``ollie.yaml``). This file should contain all
    the required preset variables for that system and the environment variables
    ``module_actions`` and ``export_vars``.

    By instancing the ``EnvironmentInfos`` class, the environment information for
    the specified model or coupled setup is compiled and stored in
    ``self.commands``. If there are environment variables inside the ``general``
    section, ``__init__`` will ignore the environment variables from the standalone
    component files, and it will define the ``general.environment_changes`` for
    each component of the setup.

    Parameters
    ----------
    run_or_compile : str
        A string indicating whether ``EnvironmentInfos`` was instanced from a
        compilation operation (``compiletime``) or a run (``runtime``).
    complete_config : dict
        Dictionary containing all the compiled information from the `yaml` files
        needed for the current `ESM-Tools` operation.
    model : string
        Model for which the environment is required. If not defined, this method
        will loop through all the available keys in ``complete_config``.
    """

    def __init__(self, run_or_compile, complete_config=None, model=None):
        # Ensure local copy of complete config to avoid mutating it... (facepalm)
        complete_config = copy.deepcopy(complete_config)
        # Load computer dictionary or initialize it from the correct machine file
        if complete_config and "computer" in complete_config:
            self.config = complete_config["computer"]
        else:
            self.machine_file = esm_parser.determine_computer_from_hostname()
            self.config = esm_parser.yaml_file_to_dict(self.machine_file)
            esm_parser.basic_choose_blocks(self.config, self.config)
            esm_parser.recursive_run_function(
                [],
                self.config,
                "atomic",
                esm_parser.find_variable,
                self.config,
                [],
                True,
            )

        # Add_s can only be inside choose_ blocks in the machine file
        for entry in ["add_module_actions", "add_export_vars", "add_unset_vars"]:
            if entry in self.config:
                del self.config[entry]

        # Load the general environments if any
        self.general_environment(complete_config, run_or_compile)

        # If the model is defined during the instantiation of the class (e.g.
        # during esm_master with a coupled setup), get the environment for that
        # model. Otherwise, loop through all the keys of the complete_config dictionary
        if model:
            self.apply_config_changes(run_or_compile, complete_config, model)
        else:
            for model in complete_config:
                self.apply_config_changes(run_or_compile, complete_config, model)

        # Add the ENVIRONMENT_SET_BY_ESMTOOLS into the exports
        self.add_esm_var()
        # Define the environment commands for the script
        self.commands = self.get_shell_commands()

    def add_esm_var(self):
        """
        Adds the ENVIRONMENT_SET_BY_ESMTOOLS=TRUE to the config, for later
        dumping to the shell script.
        """

        if "export_vars" in self.config:
            self.config["export_vars"]["ENVIRONMENT_SET_BY_ESMTOOLS"] = "TRUE"
        else:
            self.config["export_vars"] = {"ENVIRONMENT_SET_BY_ESMTOOLS": "TRUE"}

    def apply_config_changes(self, run_or_compile, config, model):
        """
        Calls ``apply_model_changes`` with the selected configuration for the
        ``model``.
        """

        self.apply_model_changes(
            model, run_or_compile=run_or_compile, modelconfig=config[model]
        )

    def apply_model_changes(self, model, run_or_compile="runtime", modelconfig=None):
        """
        Applies the ``environment_changes``, ``compiletime_environment_changes``,
        and/or ``runtime_environment_changes`` to the environment configuration of the
        ``model`` component. Note that ``model`` can be either a component (e.g.
        ``fesom``) or ``general``.

        Parameters
        ----------
        model : str
            Name of the component for which changes will apply.
        run_or_compile : str
            A string indicating whether ``EnvironmentInfos`` was instanced from a
            compilation operation (``compiletime``) or a run (``runtime``).
        modelconfig : dict
            Information compiled from the `yaml` files for this specific component.
        """

        if not modelconfig:
            print("Should not happen anymore...")
            modelconfig = esm_parser.yaml_file_to_dict(
                FUNCTION_PATH + "/" + model + "/" + model
            )

        # Merge whatever is relevant to this environment operation (either compile or
        # run) to ``environment_changes``, taking care of solving possible ``choose_``
        # blocks
        thesechanges = run_or_compile + "_environment_changes"
        if thesechanges in modelconfig:
            # kh 16.09.20 the machine name is already handled here
            # additionally handle different versions of the model (i.e.
            # choose_version...) for each machine if this is possible here in a more
            # generic way, it can be refactored
            if "choose_version" in modelconfig[thesechanges]:
                if "version" in modelconfig:
                    if (
                        modelconfig["version"]
                        in modelconfig[thesechanges]["choose_version"]
                    ):
                        for k, v in modelconfig[thesechanges]["choose_version"][
                            modelconfig["version"]
                        ].items():
                            # kh 16.09.20 move up one level and replace default
                            modelconfig[thesechanges][k] = v
                del modelconfig[thesechanges]["choose_version"]

            # Perform the merging of the environment dictionaries
            if "environment_changes" in modelconfig:
                modelconfig["environment_changes"].update(modelconfig[thesechanges])
            else:
                modelconfig["environment_changes"] = modelconfig[thesechanges]

        if "environment_changes" in modelconfig:
            for entry in ["add_module_actions", "add_export_vars", "add_unset_vars"]:
                # Initialize the environment variables
                if not entry in self.config:
                    if entry in ["add_module_actions", "add_unset_vars"]:
                        self.config[entry] = []
                    elif entry == "add_export_vars":
                        self.config[entry] = {}

                if entry == "add_export_vars":
                    # Transform any list whose name contains add_export_vars into a
                    # dictionary (machine-file export_vars are from now on always a
                    # dictionary but add_export_vars of components and setups are
                    # allowed to be lists for retro-compatibility)
                    self.turn_add_export_vars_to_dict(modelconfig, entry)

            # Merge the ``environment_changes`` into the general ``config``
            self.config.update(modelconfig["environment_changes"])
            # Change any ``choose_computer.*`` block in ``config`` to ``choose_*``
            self.remove_computer_from_choose(self.config)

            # Resolve ``choose_`` blocks
            esm_parser.basic_choose_blocks(self.config, self.config)

            # Remove the environment variables from the config
            for entry in ["add_module_actions", "add_export_vars", "add_unset_vars"]:
                if entry in self.config:
                    del self.config[entry]

    def turn_add_export_vars_to_dict(self, modelconfig, entry):
        """
        Turns the given ``entry`` in ``modelconfig`` (normally ``add_export_vars``) into
        a dictionary, if it is not a dictionary yet. This function is necessary for
        retro-compatibility of configuration files having ``add_export_vars`` defined as
        list of strings, instead of as dictionaries.

        Parameters
        ----------
        modelconfig : dict
            Information compiled from the `yaml` files for this specific component.
        entry : str
            The environment variable (originally developed for ``add_export_vars``) to
            be turned into a dictionary.
        """

        # Find the variables whose names contains the entry (e.g. add_export_vars)
        path_sep = ","
        entry_paths = esm_parser.find_key(
            modelconfig["environment_changes"],
            entry,
            paths2finds=[],
            sep=path_sep,
        )
        # Loop through the variables
        for entry_path in entry_paths:
            # Split the path and define the export_dict dictionary that links to the
            # current entry. Later, if the content of export_dict is a list it will be
            # turned into a dictionary itself
            path_to_var = entry_path.split(path_sep)
            path_to_var = [esm_parser.convert(leaf) for leaf in path_to_var]
            if len(path_to_var) > 1:
                export_dict = esm_parser.find_value_for_nested_key(
                    modelconfig["environment_changes"],
                    path_to_var[-2],
                    path_to_var[:-2],
                )
            else:
                export_dict = modelconfig["environment_changes"]
            # Get the value of export_dict
            export_vars = export_dict[path_to_var[-1]]

            # If export_vars is a list transform it into a dictionary
            if isinstance(export_vars, list):
                self.env_list_to_dict(export_dict, path_to_var[-1])

    def env_list_to_dict(self, export_dict, key):
        """
        Transforms lists in ``export_dict`` in dictionaries. This allows to add lists of
        ``export_vars`` to the machine-defined ``export_vars`` that should always be a
        dictionary. Note that lists are always added at the end of the ``export_vars``,
        if you want to edit variables of an already existing dictionary make your
        ``export_var`` be a dictionary.

        Avoids destroying repetitions of elements by adding indexes to the keys of the
        newly transformed dictionary, for example:

        .. code-block::yaml
           your_model:
               environment_changes:
                   add_export_vars:
                       - 'SOMETHING=dummy'
                       - 'somethingelse=dummy'
                       - 'SOMETHING=dummy'

        The ``export_dict[key]`` (where ``key = add_export_vars``) will be transformed
        in this function from being a list to be the following dictionary:

        .. code-block::yaml
           'SOMETHING=dummy[(0)][(list)]': 'SOMETHING=dummy'
           'somethingelse=dummy[(0)][(list)]': 'somethingelse=dummy'
           'SOMETHING=dummy[(1)][(list)]': "SOMETHING=dummy'

        Note that, once all the environments are resolved, and before writing the
        exports in the bash files, the ``export_vars`` dictionary is transformed again
        into a list and the indexes and ``[(list)]`` strings are removed.

        Parameters
        ----------
        export_dict : dict
            ``export_var`` dictionary which value is a list. This list is transformed
            into a dictionary.
        key : str
            The key to the value.
        """
        # Load the value
        export_vars = export_dict[key]
        # Check if the value is a list TODO: logging
        if not isinstance(export_vars, list):
            print(
                f"The only reason to use this function is if {key} is a list, and it "
                + "is not in this case..."
            )
            sys.exit(1)

        # Loop through the elements of the list
        new_export_vars = {}
        for var in export_vars:
            # Initialize index
            index = 0
            while True:
                # If the key with the current index already exists move the move the
                # index forward
                if var + f"[({index})][(list)]" in new_export_vars:
                    index += 1
                # If the key with the current index does not exist yet, add the element
                # to the dictionary
                else:
                    new_export_vars[f"{var}[({index})][(list)]"] = var
                    break

        # Redefined the transformed dictionary
        export_dict[key] = new_export_vars

    def general_environment(self, complete_config, run_or_compile):
        """
        Checks if there are ``environment_changes`` inside the ``general`` section, and
        if that is the case, ignore the changes loaded from the component files.

        Parameters
        ----------
        complete_config : dict
            Dictionary containing all the compiled information from the `yaml` files
            needed for the current `ESM-Tools` operation.
        run_or_compile : str
            A string indicating whether ``EnvironmentInfos`` was instanced from a
            compilation operation (``compiletime``) or a run (``runtime``).
        """

        # If the general section exists load the general environments
        general_env = False
        if "general" in complete_config:
            # Is it a coupled setup?
            coupled_setup = complete_config["general"].get("coupled_setup", False)

            # Check if a general setup environment exists that will overwrite the
            # component setups
            if coupled_setup and (
                "compiletime_environment_changes" in complete_config["general"]
                or "runtime_environment_changes" in complete_config["general"]
                or "environment_changes" in complete_config["general"]
            ):  # TODO: do this if the model include other models and the environment is
                # labelled as priority over the other models environment (OIFS case)
                general_env = True
                self.apply_config_changes(run_or_compile, complete_config, "general")

        # If there is a general environment remove all the model specific environments
        # defined in the model files and preserve only the model specific environments
        # that are explicitly defined in the setup file
        if general_env:
            self.load_component_env_changes_only_in_setup(complete_config)

    def load_component_env_changes_only_in_setup(self, complete_config):
        """
        Removes all the model specific environments defined in the component files and
        preserve only the component-specific environments that are explicitly defined in
        the setup file.

        Parameters
        ----------
        complete_config : dict
            Dictionary containing all the compiled information from the `yaml` files
            needed for the current `ESM-Tools` operation.
        """

        # Get necessary variables
        setup = complete_config.get("general", {}).get("model", None)
        version = str(complete_config.get("general", {}).get("version", None))
        models = complete_config.get("general", {}).get("models", None)
        # Check for errors TODO: logging
        if not models:
            print(
                "Use the EnvironmentInfos.load_component_env_changes_only_in_setup "
                + "method only if complete_config has a general chapter that includes "
                + "a models list"
            )
            sys.exit(1)

        # Find the setup file
        include_path, needs_load = esm_parser.look_for_file(
            setup,
            setup + "-" + version,
        )
        # If setup file not found throw and error TODO: logging
        if not include_path:
            print(f"File for {setup}-{version} not found")
            sys.exit(1)
        # Load the file TODO: logging
        if needs_load:
            setup_config = esm_parser.yaml_file_to_dict(include_path)
        else:
            print(f"A setup needs to load a file so this line shouldn't be reached")
            sys.exit(1)

        # Add the attachment files (e.g. the environment variables can be in a
        # further_reading file)
        for attachment in esm_parser.CONFIGS_TO_ALWAYS_ATTACH_AND_REMOVE:
            # Add the attachment file chapters (e.g. there is a further_reading chapter
            # at the same level of general and the components)
            esm_parser.attach_to_config_and_remove(setup_config, attachment)
            # Add the attachment files in each chapter (i.e. in general, components,
            # etc.)
            for component in list(setup_config):
                esm_parser.attach_to_config_and_remove(
                    setup_config[component],
                    attachment,
                )

        # Define the possible environment variables
        environment_vars = [
            "environment_changes",
            "compiletime_environment_changes",
            "runtime_environment_changes",
        ]
        # Loop through the models
        for model in models:
            # Sanity check TODO: logging
            if model not in complete_config:
                print(f"The chapter {model} does not exist in complete_config")
                sys.exit(1)
            # Load the configuration of this model
            model_config = complete_config[model]
            # Loop through the possible environment variables
            for env_var in environment_vars:
                # If the environment variable exists replace it with the one defined in
                # the setup file for that model:
                # 1. Delete the variable
                if env_var in model_config:
                    del model_config[env_var]
                # 2. Redefine the variable
                if env_var in setup_config.get(model, {}):
                    # Solve any unresolved variables in the reloaded setup environment
                    # TODO: change this to  be out of the loop using the method
                    # ``model_config.finalize()``, currently not working due to
                    # a problem with the dates
                    esm_parser.recursive_run_function(
                        [],
                        setup_config[model][env_var],
                        "atomic",
                        esm_parser.find_variable,
                        complete_config,
                        {},
                        {},
                    )
                    # Actually redefine the variable
                    model_config[env_var] = setup_config[model][env_var]

    def replace_model_dir(self, model_dir):
        """
        Replaces any instances of ${model_dir} in the config section
        "export_vars" with the argument

        Parameters
        ----------
        model_dir : str
            The replacement string for ${model_dir}
        """
        for entry in ["export_vars"]:
            if entry in self.config:
                newlist = []
                for line in self.config[entry]:
                    newline = line.replace("${model_dir}", model_dir)
                    newlist.append(newline)
                self.config[entry] = newlist

    def get_shell_commands(self):
        """
        Gathers module actions and export variables from the config to a list,
        prepending appropriate shell command words (e.g. module and export).

        If the ``export_vars`` dictionary contains variables with repetition
        indexes (``[(int)]``) or ``[(list)]``, those are removed before returning the
        command list.

        Returns
        -------
        environment : list
            A list of the environment operations, to be used in the compilation and run
            scripts.
        """

        environment = []
        # Write module actions
        if "module_actions" in self.config:
            for action in self.config["module_actions"]:
                # seb-wahl: workaround to allow source ... to be added to the batch header
                # until a proper solution is available. Required with FOCI
                if action.startswith("source"):
                    environment.append(action)
                else:
                    environment.append(f"module {action}")
        # Add an empty string as a newline:
        environment.append("")
        if "export_vars" in self.config:
            for var in self.config["export_vars"]:
                # If export_vars is a dictionary
                if isinstance(self.config["export_vars"], dict):
                    # If the variable is a dictionary itself (e.g. "AWI_FESOM_YAML"
                    # in fesom-1.4) add the contents of the dictionary as the value of
                    # the exported variable inside '""'
                    if isinstance(self.config["export_vars"][var], dict):
                        key = var
                        value = self.config["export_vars"][key]
                        environment.append(f"export {key}='{str(value)}'")
                    # If the value of the variable is not a dictionary
                    else:
                        key = var
                        value = self.config["export_vars"][key]
                        # Define the pattern for indexes [(int)]
                        ipattern = "\[+\(\d+\)+\]$"
                        # If the variable was added as a list produce the correct string
                        if key.endswith("[(list)]"):
                            key = key.replace("[(list)]", "")
                            environment.append(f"export {value}")
                        # If the variable contained a repetition index, remove it
                        elif re.search(ipattern, key):
                            environment.append(
                                f"export {re.sub(ipattern, '', key)}={str(value)}"
                            )
                        # It it is a normal variable return the export command
                        else:
                            environment.append(f"export {key}={str(value)}")
                # If export_vars is a list append the export command (this should not
                # happen anymore as the export_vars in the machine files should be all
                # defined now as dictionaries
                else:
                    environment.append("export {str(var)}")
        environment.append("")
        # Write the unset commands
        if "unset_vars" in self.config:
            for var in self.config["unset_vars"]:
                environment.append(f"unset {var}")

        return environment

    def write_dummy_script(self, include_set_e=True):
        """
        Writes a dummy script containing only the header information, module
        commands, and export variables. The actual compile/configure commands
        are added later.

        Parameters
        ----------
        include_set_e : bool
            Default to True, whether or not to include a ``set -e`` at the
            beginning of the script. This causes the shell to stop as soon as
            an error is encountered.
        """
        # Check for sh_interpreter
        if "sh_interpreter" not in self.config:
            print('WARNING: "sh_interpreter" not defined in the machine yaml')
        with open("dummy_script.sh", "w") as script_file:
            # Write the file headings
            script_file.write(
                f'#!{self.config.get("sh_interpreter", "/bin/bash")} -l\n'
            )
            script_file.write(
                "# Dummy script generated by esm-tools, to be removed later: \n"
            )
            if include_set_e:
                script_file.write("set -e\n")

            # Write the module and export commands
            for command in self.commands:
                script_file.write(f"{command}\n")
            script_file.write("\n")

    def remove_computer_from_choose(self, chapter):
        """
        Recursively remove ``computer.`` from all the `choose_` keys.

        Parameters
        ----------
        chapter : dict
            Dictionary to search for ``choose_computer.`` blocks.
        """
        all_keys = list(chapter.keys())
        for key in all_keys:
            if isinstance(key, str) and "choose_computer." in key:
                newkey = key.replace("computer.", "")
                chapter[newkey] = chapter[key]
                del chapter[key]
                key = newkey
            if isinstance(chapter[key], dict):
                self.remove_computer_from_choose(chapter[key])

    @staticmethod
    def cleanup_dummy_script():
        """
        Removes the ``dummy_script.sh`` if it exists.
        """
        try:
            os.remove("dummy_script.sh")
        except OSError:
            print("No file dummy_script.sh there; nothing to do...")

    @staticmethod
    def add_commands(commands, name):
        """
        Writes all commands in a list to a file named ``<name>_script.sh``,
        located in the current working directory. The header from this script
        is read from ``dummy_script.sh``, also in the current working
        directory.

        Parameters
        ----------
        commands : list of str
            List of the commands to write to the file after the header
        name : str
            Name of the script, generally something like ``comp_echam-6.3.05``

        Returns
        -------
        str :
            ``name`` + "_script.sh"
        """
        if commands:
            with open(f"{name}_script.sh", "w") as newfile:
                with open("dummy_script.sh", "r") as dummy_file:
                    newfile.write(dummy_file.read())
                for command in commands:
                    newfile.write(f"{command}\n")
        return f"{name}_script.sh"

    def output(self):
        esm_parser.pprint_config(self.config)


class environment_infos(EnvironmentInfos):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Please change your code to use EnvironmentInfos!",
            DeprecationWarning,
            stacklevel=2,
        )
        super(environment_infos, self).__init__(*args, **kwargs)



"""
Contains functions for dealing with PBS-based batch systems
"""
import os
import subprocess
import sys
import esm_parser


class Pbs:
    """
    Deals with PBS, allowing you to check if a job is submitted, get the current job
    ID, generate the job launcher commands (e.g. ``aprun``), get the current job
    state, and check if a job is still running. Note: there is not an equivalent file
    for the slurm hostfile, because all the job information is contained in the
    launcher command.


    Parameters
    ----------
    config : dict
        Configuration dictionary containing information about the experiment and
        experiment directory.
    """

    def __init__(self, config):
        # No hostfile for PBS
        self.filename = ""
        self.path = ""

    @staticmethod
    def check_if_submitted():
        """
        Determines if a job is submitted in the currently running shell by checking for
        ``PBS_JOBID`` variable in the environment.

        Returns
        -------
        bool
            Boolean standing for submission state.
        """
        return "PBS_JOBID" in os.environ

    @staticmethod
    def get_jobid():
        """
        Gets the current ``PBS_JOBID``.

        Returns
        -------
        str or None
            If ``PBS_JOBID`` is defined returns the jobid as a string.
        """
        return os.environ.get("PBS_JOBID")

    @staticmethod
    def calc_launcher_flags(config, model, cluster):
        """
        Calculates the launcher flags for the job luncher based on the ``nproc`` of the
        different components, with the possibility of using heterogeneous
        parallelization (using both MPI and OpenMPI).

        Launcher flags will vary from job launcher to job launcher, and therefore, a
        variable ``launcher_flags`` needs to be defined for each different launcher in
        the ``pbs.yaml`` configuration file. This variable should contain the flags
        themselves followed by their respective values to be calculated in this method
        and substituted in their respective tags of the ``launcher_flags`` variable.
        This method supports the following variables:

            - ``@nproc@``: total number of MPI tasks for the component taken from
              ``<component>.nproc``.
            - ``@nproc_per_node@``: number of MPI tasks per node calculated from
              the number of number of CPUs per MPI task (``<component>.cpus_per_proc``),
              and the number of cores per node in the machine
              (``computer.cores_per_node``).
            - ``@cpus_per_proc@``: number of CPUs per MPI task to be used
              (``<component>.cpus_per_proc``). If not defined, it takes the value of
              ``<component>.omp_num_threads``. This variables are not necessarily the
              same, to allow the use of  more cores per MPI task than OpenMP threads,
              if necessary.
            - ``@omp_num_threads@``: number of OpenMP threads per MPI task
              (``<component>.omp_num_threads``). Its default value is ``1``, if not
              defined. The number of OpenMP threads per task cannot be larger than the
              number of CPUs per task.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        model : str
            Component for which the flags are to be calculated.
        cluster : str

        Returns
        -------
        launcher_flags : str
            Launcher flags string with the calculated numbers for the ``model`` already
            substituted in the tags.
        """
        launcher = config["computer"]["launcher"]
        launcher_flags = config["computer"]["launcher_flags"]
        if "nproc" in config[model]:
            # aprun flags commented following the conventions in p. 14 of the ALEPH ppt
            # manual day_1.session_2.advanced_use_of_aprun.ppt

            # Total number of PEs (MPI-ranks) (e.g. aprun -n)
            nproc = config[model]["nproc"]
            # Cores per node
            # cores_per_node = config["computer"]["cores_per_node"]
            if cluster == "compute":
                cores_per_node = config["computer"]["partitions"]["compute"][
                    "cores_per_node"
                ]
            else:
                cores_per_node = config["computer"]["partitions"]["pp"][
                    "cores_per_node"
                ]
            # Get the OMP number of threads
            omp_num_threads = config[model].get("omp_num_threads", 1)
            # CPUs per MPI-rank (e.g. aprun -d)n
            cpus_per_proc = config[model].get("cpus_per_proc", omp_num_threads)
            # Check for CPUs and OpenMP threads
            if omp_num_threads > cpus_per_proc:
                esm_parser.user_error(
                    "OpenMP configuration",
                    (
                        "The number of OpenMP threads cannot be larger than the number"
                        + "of CPUs per MPI task requested. Your values:\n"
                        + f"    {model}.omp_num_threads: {omp_num_threads}\n"
                        + f"    {model}.cpus_per_proc: {cpus_per_proc}\n"
                    ),
                )
            # Number of nodes needed
            nodes = int(nproc * cpus_per_proc / cores_per_node) + (
                (nproc * cpus_per_proc) % cores_per_node > 0
            )
            # PEs (MPI-ranks) per compute node (e.g. aprun -N)
            nproc_per_node = int(nproc / nodes)
        elif "nproca" in config[model] and "procb" in config[model]:
            esm_parser.user_error(
                "nproc", "nproca and nprocb not supported yet for pbs"
            )

        # Replace tags in the laucher flags
        replacement_tags = [
            ("@nproc@", nproc),
            ("@nproc_per_node@", nproc_per_node),
            ("@cpus_per_proc@", cpus_per_proc),
            ("@omp_num_threads@", omp_num_threads),
        ]
        # Replace all tags
        for (tag, repl) in replacement_tags:
            launcher_flags = launcher_flags.replace(tag, str(repl))
        # Substitute @MODEL@ with the model name
        launcher_flags = launcher_flags.replace("@MODEL@", model.upper())

        return launcher_flags

    def prepare_launcher(self, config, cluster):
        """
        Loops through the components to generate job launcher flags and execution
        commands, to be appended in substitution to the ``@components@`` tag, in
        ``computer.execution_command``, that would later be used in the writing of the
        ``.run`` file, in ``batch_system.py``.

        .. Note: PBS does not support yet multi_apruns

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        cluster : str
        """
        # PBS does not support yet multi_apruns
        # if config['general'].get('multi_apruns'):
        #    self.calc_requirements_multi_aprun(config)
        #    return

        component_lines = []
        # Read in the separator to be used in between component calls in the job
        # launcher
        sep = config["computer"].get("launcher_comp_sep", "\\\n    ") + " "
        # Loop through the components
        for model in config["general"]["valid_model_names"]:
            command = None
            # Read in execution command
            if "execution_command" in config[model]:
                command = config[model]["execution_command"]
            elif "executable" in config[model]:
                command = config[model]["executable"]
            # Prepare the MPMD commands
            if command:
                launcher = config["computer"].get("launcher")
                launcher_flags = self.calc_launcher_flags(config, model, cluster)
                component_lines.append(f"{launcher_flags} ./{command} ")

        # Merge each component flags and commands into a single string
        components = sep.join(component_lines)
        # Substitute the ``@components@`` tag for the final launch command
        config["computer"]["execution_command"] = config["computer"][
            "execution_command"
        ].replace("@components@", components)

    @staticmethod
    def add_pre_launcher_lines(config, runfile):
        """
        Adds pre-launcher lines to the ``runfile``.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        runfile : io.TextIOWrapper
            File wrapper object for writing of the lines
            (``runfile.write("<your_line_here>")``).
        """
        # This changes the name of the output stream to include the $PBS_JOBID. This
        # cannot be done in the header because PBS does not support its own variables
        # to be used there (at least in the ALEPH's version).
        #runfile.write(f'qalter $PBS_JOBID -o {config["computer"]["thisrun_logfile"]}\n')
        pass

    @staticmethod
    def write_het_par_wrappers(config):
        if config["general"].get("verbose", False):
            print(
                f"Skipping the het-par wrapper as it is not needed for {config['computer']['batch_system']}"
            )
        return config

    @staticmethod
    def get_job_state(jobid):
        """
        Returns the jobstate. See ``man qstat`` for more details.

        Parameters
        ----------
        jobid : str
            The PBS job id as displayed in, e.g. ``qstat``.

        Returns
        -------
        str or None
            The short job state.
        """
        state_command = f"qstat {str(jobid)}"

        qstat_output = subprocess.Popen(
            state_command.split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()[0]
        qstat_split = str(qstat_output).split()
        if len(qstat_split) > 2:
            return qstat_split[-3]

    @staticmethod
    def job_is_still_running(jobid):
        """
        Returns ``True`` if the job is still running or queueing.

        Parameters
        ----------
        jobid : str
            The PBS job id as displayed in, e.g. ``qstat``.

        Returns
        -------
        bool
            Boolean standing for a job running or queueing.
        """
        return bool(Pbs.get_job_state(jobid))



# -*- coding: utf-8 -*-

"""
Database module, including the SQLAlchemy database object and DB-related
utilities.
"""
from .compat import basestring
from .extensions import db

# Alias common SQLAlchemy names
Column = db.Column
relationship = db.relationship


class CRUDMixin(object):
    """Mixin that adds convenience methods for CRUD (create, read, update,
    delete) operations."""

    @classmethod
    def create(cls, **kwargs):
        """Create a new record and save it the database."""
        instance = cls(**kwargs)
        return instance.save()

    def update(self, commit=True, **kwargs):
        """Update specific fields of a record."""
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        return commit and self.save() or self

    def save(self, commit=True):
        """Save the record."""
        db.session.add(self)
        if commit:
            db.session.commit()
        return self

    def delete(self, commit=True):
        """Remove the record from the database."""
        db.session.delete(self)
        return commit and db.session.commit()


class Model(CRUDMixin, db.Model):
    """Base model class that includes CRUD convenience methods."""

    __abstract__ = True


# From Mike Bayer's "Building the app" talk
# https://speakerdeck.com/zzzeek/building-the-app
class SurrogatePK(object):
    """A mixin that adds a surrogate integer 'primary key' column named ``id`` to any declarative-mapped class."""

    __table_args__ = {"extend_existing": True}

    id = Column(db.Integer, primary_key=True)

    @classmethod
    def get_by_id(cls, record_id):
        """Get record by ID."""
        if any(
            (
                isinstance(record_id, basestring) and record_id.isdigit(),
                isinstance(record_id, (int, float)),
            )
        ):
            return cls.query.get(int(record_id))
        return None


def reference_col(
    tablename, nullable=False, pk_name="id", foreign_key_kwargs=None, column_kwargs=None
):
    """Column that adds primary key foreign key reference.
    Usage: ::
        category_id = reference_col('category')
        category = relationship('Category', backref='categories')
    """
    foreign_key_kwargs = foreign_key_kwargs or {}
    column_kwargs = column_kwargs or {}

    return Column(
        db.ForeignKey(f"{tablename}.{pk_name}", **foreign_key_kwargs),
        nullable=nullable,
        **column_kwargs,
    )



import sys
import os
import six

from . import batch_system
from . import logfiles
from . import helpers
from . import chunky_parts
from . import workflow
import esm_parser


def submit(config):
    if config["general"]["verbose"]:
        six.print_("\n", 40 * "+ ")
    print("Submitting jobscript to batch system...")
    print()
    print(f"Output written by {config['computer']['batch_system']}:")
    if config["general"]["verbose"]:
        six.print_("\n", 40 * "+ ")
        for command in config["general"]["submit_command"]:
            print(command)
    for command in config["general"]["submit_command"]:
        os.system(command)
    return config


def resubmit_batch_or_shell(config, batch_or_shell, cluster=None):
    config = config["general"]["batch"].write_simple_runscript(
        config, cluster, batch_or_shell
    )
    if not check_if_check(config):
        config = submit(config)
    return config


def resubmit_SimulationSetup(config, cluster=None):
    monitor_file = logfiles.logfile_handle
    # Jobs that should be started directly from the compute job:

    jobtype = config["general"]["jobtype"]

    monitor_file.write(f"{cluster} for this run:\n")
    command_line_config = config["general"]["command_line_config"]
    command_line_config["jobtype"] = cluster

    monitor_file.write(f"Initializing {cluster} object with:\n")
    monitor_file.write(str(command_line_config))
    # NOTE(PG) Non top level import to avoid circular dependency:

    os.chdir(config["general"]["started_from"])
    from .sim_objects import SimulationSetup

    cluster_obj = SimulationSetup(command_line_config)

    monitor_file.write(f"{cluster} object built....\n")

    if f"{cluster}_update_{jobtype}_config_before_resubmit" in cluster_obj.config:
        monitor_file.write(
            f"{cluster} object needs to update the calling job config:\n"
        )
        # FIXME(PG): This might need to be a deep update...?
        config.update(
            cluster_obj.config[f"{cluster}_update_{jobtype}_config_before_resubmit"]
        )

    if not check_if_check(config):

        monitor_file.write(f"Calling {cluster} job:\n")
        config["general"]["experiment_over"] = cluster_obj(kill_after_submit=False)

    return config


def get_submission_type(cluster, config):
    # Figure out if next job is resubmitted to batch system,
    # just executed in shell or invoked as new SimulationSetup
    # object

    clusterconf = config["general"]["workflow"]["subjob_clusters"][cluster]

    if clusterconf.get("submit_to_batch_system", False):
        submission_type = "batch"
    elif cluster in ["newrun", "prepcompute", "tidy", "inspect", "viz"]:
        submission_type = "SimulationSetup"
    else:
        submission_type = "shell"

    return submission_type


def end_of_experiment(config):
    if config["general"]["next_date"] >= config["general"]["final_date"]:
        monitor_file = logfiles.logfile_handle
        monitor_file.write("Reached the end of the simulation, quitting...\n")
        config["general"]["experiment_over"] = True
        helpers.write_to_log(config, ["# Experiment over"], message_sep="")
        return True
    return False


def end_of_experiment_all_models(config):
    index = 1
    expid = config["general"]["expid"]
    while "model" + str(index) in config["general"]["original_config"]:
        if (
            not config["model" + str(index)]["setup_name"]
            == config["general"]["setup_name"]
        ):
            experiment_done = False
            setup_name = config["model" + str(index)]["setup_name"]
            print(f"Testing if {setup_name} is already done...")
            logfile = (
                config["general"]["experiment_log_dir"]
                + "/"
                + expid
                + "_"
                + setup_name
                + ".log"
            )
            if os.path.isfile(logfile):
                with open(logfile, "r") as open_logfile:
                    logfile_array = open_logfile.readlines()
                    for line in logfile_array:
                        if "# Experiment over" in line:
                            print(f"    ...{setup_name} is done.")
                            experiment_done = True
                            break
            if not experiment_done:
                print("Still something left to do...")
                return False
        index += 1
    print("Nothing left to do...")
    return True


def check_if_check(config):
    if config["general"]["check"]:
        print(
            "Actually not submitting anything, this job preparation was launched in 'check' mode (-c)."
        )
        print()
        return True
    else:
        return False


def maybe_resubmit(config):

    jobtype = config["general"]["jobtype"]

    nextrun = resubmit_recursively(config, jobtype=jobtype)

    if nextrun:  # submit list contains stuff from next run

        config = _increment_date_and_run_number(config)
        config = _write_date_file(config)

        if end_of_experiment(config):
            if config["general"].get("iterative_coupling", False):
                if end_of_experiment_all_models(config):
                    return config
            else:
                # config = chunky_parts._update_chunk_date_file(config)
                return config

        cluster = config["general"]["workflow"]["first_task_in_queue"]
        nextrun = resubmit_recursively(
            config, list_of_clusters=[cluster], nextrun_in=True
        )

    return config


def resubmit_recursively(config, jobtype=None, list_of_clusters=None, nextrun_in=False):
    nextrun = False

    if not list_of_clusters:
        list_of_clusters = config["general"]["workflow"]["subjob_clusters"][
            jobtype
        ].get("next_submit", [])

    for cluster in list_of_clusters:
        if (
            cluster == config["general"]["workflow"]["first_task_in_queue"]
            and not nextrun_in
        ):
            nextrun = True
        else:
            if not workflow.skip_cluster(cluster, config):
                submission_type = get_submission_type(cluster, config)
                if submission_type == "SimulationSetup":
                    resubmit_SimulationSetup(config, cluster)
                elif submission_type in ["batch", "shell"]:
                    resubmit_batch_or_shell(config, submission_type, cluster)
            else:
                print(f"Skipping {cluster}")
                nextrun = (
                    resubmit_recursively(config, jobtype=cluster, nextrun_in=nextrun_in)
                    or nextrun
                )
    return nextrun


def _increment_date_and_run_number(config):
    config["general"]["run_number"] += 1
    config["general"]["current_date"] += config["general"]["delta_date"]

    config["general"]["command_line_config"]["current_date"] = config["general"][
        "current_date"
    ].format(form=9, givenph=False, givenpm=False, givenps=False)

    config["general"]["command_line_config"]["run_number"] = config["general"][
        "run_number"
    ]

    config = chunky_parts.update_command_line_config(config)

    return config


def _write_date_file(config):  # self, date_file=None):
    # monitor_file = config["general"]["logfile"]
    monitor_file = logfiles.logfile_handle

    # if not date_file:
    date_file = (
        f"{config['general']['experiment_scripts_dir']}"
        f"/{config['general']['expid']}_{config['general']['setup_name']}.date"
    )

    with open(date_file, "w") as date_file:
        date_file.write(
            config["general"]["current_date"].output()
            + " "
            + str(config["general"]["run_number"])
        )
    monitor_file.write("writing date file \n")
    return config



def prepare_environment(config):



    environment_dict = {
            "TEST_IN_ENV": "testvar_in_env_exported_couple_in_vilma",
            "ICE_TO_VILMA": 1,
            "VILMA_GRID_input": config["vilma"]["grid_input"],
            "COUPLE_DIR": config["general"]["experiment_couple_dir"],
            "solidearth_ice_thickness_file":(
                config["general"]["experiment_couple_dir"] +
                "/ice_thickness.nc"
                ),
            "ADD_UNCHANGED_ICE": config["vilma"].get("add_unchanged_ice", False),
            "EISLASTFILE_vilma":  (
                config["vilma"]["experiment_input_dir"] +
                "/" +
                config["vilma"]["eislastfile"]
                ),
            "RUN_NUMBER_vilma": config["general"]["run_number"],
            "RUN_DATE_STAMP": config["general"]["run_datestamp"],
            "LAST_RUN_DATE_STAMP": config["general"]["last_run_datestamp"],
            "DATA_DIR_vilma": config["vilma"]["experiment_restart_out_dir"],
            "INITIAL_YEAR_vilma": config["general"]["initial_date"].syear,
            "NYEAR_vilma_standalone": config["general"]["nyear"],
            "FINAL_YEAR_vilma": config["general"]["final_date"].syear,
            "EISLASTCONF_vilma":(
                config["vilma"]["experiment_config_dir"] +
                "/inp/" +
                config["vilma"]["eislastconf"]
                )

            }
    
    if environment_dict["ADD_UNCHANGED_ICE"] == False:
        environment_dict["ADD_UNCHANGED_ICE"] = 0
    elif environment_dict["ADD_UNCHANGED_ICE"] == True:
        environment_dict["ADD_UNCHANGED_ICE"] = 1


    print(environment_dict)
    return environment_dict







import os
import shutil
import subprocess
import copy
import pathlib

import f90nml
import six
import yaml
from colorama import Fore, Back, Style, init

import esm_tools
import esm_calendar
import esm_parser
import esm_rcfile
import esm_runscripts

from .batch_system import batch_system
from .filelists import copy_files, log_used_files
from .helpers import end_it_all, evaluate, write_to_log
from .namelists import Namelist
from loguru import logger

#####################################################################
#                                   compute jobs                    #
#####################################################################


def run_job(config):
    config["general"]["relevant_filetypes"] = [
        "bin",
        "config",
        "forcing",
        "input",
        "restart_in",
    ]
    config = evaluate(config, "compute", "compute_recipe")
    return config


def compile_model(config):
    """Compiles the desired model before the run starts"""
    model = config["general"]["setup_name"]
    version = config["general"].get("version") or config[model].get("version")
    if not version:
        return config
    if config.get("general", {}).get("run_number") == 1:
        print("First year, checking if we need to compile...")
        if not config.get("general", {}).get("use_compiled_model", True):
            print(f"Huhu --> compiling {model}-{version}")
            subprocess.run(
                f"esm_master install-{model}-{version}",
                shell=True,
                cwd=config["general"]["experiment_src_dir"],
            )
            config["general"]["model_dir"] = (
                config["general"]["experiment_src_dir"] + f"/{model}-{version}"
            )
    return config


def all_files_to_copy_append(
    config, model, filetype, categ, file_source, file_interm, file_target
):
    if file_source:
        if not filetype + "_sources" in config[model]:
            config[model][filetype + "_sources"] = {}
        config[model][filetype + "_sources"][categ] = file_source
    if file_interm:
        if not filetype + "_intermediate" in config[model]:
            config[model][filetype + "_intermediate"] = {}
        config[model][filetype + "_intermediate"][categ] = file_interm
    if file_target:
        if not filetype + "_targets" in config[model]:
            config[model][filetype + "_targets"] = {}
        config[model][filetype + "_targets"][categ] = file_target

    if filetype + "_files" in config[model]:
        config[model][filetype + "_files"][categ] = categ

    return config


def prepare_coupler_files(config):
    if config["general"]["standalone"] is False:
        coupler_filename = config["general"]["coupler"].prepare(
            config, config["general"]["coupler_config_dir"]
        )
        coupler_name = config["general"]["coupler"].name
        if coupler_name == "yac":
            couplingfile = "coupling.xml"
        else:
            couplingfile = "namcouple"

        all_files_to_copy_append(
            config,
            coupler_name,
            "config",
            couplingfile,
            config["general"]["coupler_config_dir"] + "/" + coupler_filename,
            None,
            None,
        )
    return config


def create_new_files(config):
    for model in list(config):
        for filetype in config["general"]["all_filetypes"]:
            if "create_" + filetype in config[model]:
                filenames = config[model]["create_" + filetype].keys()
                for filename in filenames:
                    with open(
                        config[model]["thisrun_" + filetype + "_dir"] + "/" + filename,
                        "w",
                    ) as createfile:
                        actionlist = config[model]["create_" + filetype][filename]
                        for action in actionlist:
                            if "<--append--" in action:
                                appendtext = action.replace("<--append--", "")
                                createfile.write(appendtext.strip() + "\n")
                    all_files_to_copy_append(
                        config,
                        model,
                        filetype,
                        filename,
                        config[model]["thisrun_" + filetype + "_dir"] + "/" + filename,
                        None,
                        None,
                    )
    return config


def modify_files(config):
    # for model in config:
    #     for filetype in config["general"]["all_model_filetypes"]:
    #         if filetype == "restart":
    #             nothing = "nothing"
    return config


def modify_namelists(config):
    # Load and modify namelists:

    if config["general"]["verbose"]:
        six.print_("\n" "- Setting up namelists for this run...")
        for index, model in enumerate(config["general"]["valid_model_names"]):
            print(f'{index+1}) {config[model]["model"]}')
        print()

    for model in config["general"]["valid_model_names"]:
        config[model] = Namelist.nmls_load(config[model])
        config[model] = Namelist.nmls_remove(config[model])
        if model == "echam":
            config = Namelist.apply_echam_disturbance(config)
        config[model] = Namelist.nmls_modify(config[model])
        config[model] = Namelist.nmls_finalize(
            config[model], config["general"]["verbose"]
        )

    if config["general"]["verbose"]:
        print("::: end of namelist section\n")
    return config


def copy_files_to_thisrun(config):
    if config["general"]["verbose"]:
        six.print_("PREPARING EXPERIMENT")
        # Copy files:
        six.print_("\n" "- File lists populated, proceeding with copy...")
        six.print_("- Note that you can see your file lists in the config folder")
        six.print_("- You will be informed about missing files")

    log_used_files(config)

    config = copy_files(
        config, config["general"]["in_filetypes"], source="init", target="thisrun"
    )
    return config


def copy_files_to_work(config):
    if config["general"]["verbose"]:
        six.print_("PREPARING WORK FOLDER")
    config = copy_files(
        config, config["general"]["in_filetypes"], source="thisrun", target="work"
    )
    return config


def _create_folders(config, filetypes):
    """
    Generates the experiment file tree. Foldres are created for every filetype
    except for "ignore".
    """
    for filetype in filetypes:
        if not filetype == "ignore":
            if not filetype == "work":
                if not os.path.exists(config["experiment_" + filetype + "_dir"]):
                    os.makedirs(config["experiment_" + filetype + "_dir"])
            if not os.path.exists(config["thisrun_" + filetype + "_dir"]):
                os.makedirs(config["thisrun_" + filetype + "_dir"])


def _create_setup_folders(config):
    """
    Creates all folders for an experiment. This is the plugin function called
    during the recipe. See also ``_create_folders`` for actual creation

    This also creates a small marker file at the top of
    the experiment so that the "root" can be found from inside.
    """
    _create_folders(config["general"], config["general"]["all_filetypes"])
    with open(
        config["general"]["experiment_dir"] + "/.top_of_exp_tree", "w"
    ) as top_marker:
        top_marker.write(f"Top of experiment {config['general']['expid']}")
    return config


def _create_component_folders(config):
    for component in config["general"]["valid_model_names"]:
        _create_folders(config[component], config["general"]["all_model_filetypes"])
    return config


def initialize_experiment_logfile(config):
    """
    Initializes the log file for the entire experiment.

    Creates a file ``${BASE_DIR}/${EXPID}/log/${EXPID}_${setup_name}.log``
    to keep track of start/stop times, job id numbers, and so on. Use the
    function ``write_to_log`` to put information in this file afterwards.

    The user can specify ``experiment_log_file`` under the ``general``
    section of the configuration to override the default name. Timestamps
    for each message are given by the section
    ``experiment_log_file_dateformat``, or defaults to ``Tue Mar 17
    09:36:38 2020``, i.e. ``"%c"``. Please use ``stftime`` compatable
    formats, as described here: https://strftime.org

    Parameters
    ----------
    dict :
        The experiment configuration

    Return
    ------
    dict :
        As per convention for the plug-in system; this gives back the
        entire config.

    Attention
    ---------
        Calling this has some filesystem side effects. If the run number in
        the general configuration is set to 1, and a file exists for
        ``general.exp_log_file``; this file is removed; and re-initialized.
    """

    if config["general"]["run_number"] == 1:
        if os.path.isfile(config["general"]["experiment_log_file"]):
            os.remove(config["general"]["experiment_log_file"])

        log_msg = f"# Beginning of Experiment {config['general']['expid']}"
        write_to_log(config, [log_msg], message_sep="")

    write_to_log(
        config,
        [
            str(config["general"]["jobtype"]),
            str(config["general"]["run_number"]),
            str(config["general"]["current_date"]),
            str(config["general"]["jobid"]),
            "- submitted",
        ],
    )

    # Write trace-log file now that we know where to do that
    if "trace_sink" in dir(logger):
        logfile_path = (
            f"{config['general']['experiment_dir']}/log"
            f"/{config['general']['expid']}_esm_runscripts_"
            f"{config['general']['run_datestamp']}.log"
        )

        logger.trace_sink.def_path(logfile_path)

    return config


def _write_finalized_config(config):
    """Writes <expid>_finished_config.yaml file
    Parameters
    ----------
    config : esm-tools config object
    """
    # first define the representers for the non-built-in types, as recommended
    # here: https://pyyaml.org/wiki/PyYAMLDocumentation
    def date_representer(dumper, date):
        return dumper.represent_str(f"{date.output()}")

    def calendar_representer(dumper, calendar):
        # Calendar has a __str__ method
        return dumper.represent_str(str(calendar))

    def batch_system_representer(dumper, batch_system):
        return dumper.represent_str(f"{batch_system.name}")

    def coupler_representer(dumper, coupler):
        # prevent dumping of whole namcouple
        return dumper.represent_str(f"{coupler.name}")

    def oasis_representer(dumper, oasis):
        return dumper.represent_str(f"{oasis.name}")

    def namelist_representer(dumper, f90nml):
        return dumper.represent_str(f"f90nml.name")

    # dumper object for the ESM-Tools configuration
    class EsmConfigDumper(yaml.dumper.Dumper):
        pass

    # pyyaml does not support tuple and prints !!python/tuple
    EsmConfigDumper.add_representer(
        tuple, yaml.representer.SafeRepresenter.represent_list
    )

    # Determine how non-built-in types will be printed be the YAML dumper
    EsmConfigDumper.add_representer(esm_calendar.Date, date_representer)

    EsmConfigDumper.add_representer(
        esm_calendar.esm_calendar.Calendar, calendar_representer
    )
    # yaml.representer.SafeRepresenter.represent_str)

    EsmConfigDumper.add_representer(
        esm_parser.esm_parser.ConfigSetup,
        yaml.representer.SafeRepresenter.represent_dict,
    )

    EsmConfigDumper.add_representer(batch_system, batch_system_representer)

    # format for the other ESM data structures
    EsmConfigDumper.add_representer(
        esm_rcfile.esm_rcfile.EsmToolsDir,
        yaml.representer.SafeRepresenter.represent_str,
    )

    EsmConfigDumper.add_representer(
        esm_runscripts.coupler.coupler_class, coupler_representer
    )

    EsmConfigDumper.add_representer(
        f90nml.namelist.Namelist, namelist_representer
    )

    if "oasis3mct" in config:
        EsmConfigDumper.add_representer(esm_runscripts.oasis.oasis, oasis_representer)

    config_file_path = (
        f"{config['general']['thisrun_config_dir']}"
        f"/{config['general']['expid']}_finished_config.yaml"
    )
    with open(config_file_path, "w") as config_file:
        # Avoid saving ``prev_run`` information in the config file
        config_final = copy.deepcopy(config)  # PrevRunInfo
        del config_final["prev_run"]  # PrevRunInfo

        out = yaml.dump(
            config_final, Dumper=EsmConfigDumper, width=10000, indent=4
        )  # PrevRunInfo
        config_file.write(out)
    return config


def color_diff(diff):
    for line in diff:
        if line.startswith("+"):
            yield Fore.GREEN + line + Fore.RESET
        elif line.startswith("-"):
            yield Fore.RED + line + Fore.RESET
        elif line.startswith("^"):
            yield Fore.BLUE + line + Fore.RESET
        else:
            yield line


def update_runscript(fromdir, scriptsdir, tfile, gconfig, file_type):
    """
    Updates the script ``tfile`` in the directory ``scriptdir`` with the file in
    the directory ``fromdir`` if the update flag (``-U``) is used during the call of
    ``esm_runscripts``. If that flag is not used and the source and target are different
    then raises a user-friendly error recommending to use the ``-U`` flag with the warning
    that the files will be overwritten.
    Parameters
    ----------
    cls : obj
        Compute object.
    fromdir : str
        Path of the source.
    scriptsdir : str
        Path of the target.
    tfile : str
        Name of the script to be updated.
    gconfig : dict
        Dictionary containing the general information about the compute task.
    file_type : str
        String specifying the nature of the file, only necessary for printing information
        and for the error description.
    Exceptions
    ----------
    UserError
        If the target and source are different and the ``-U`` flag is not used when calling
        ``esm_runscripts``, returns an error.
    """

    # if `tfile` contains a full path of the runscript then remove the leading path
    tfile = os.path.basename(tfile)

    # If the target file in ``scriptsdir`` does not exist, then copy the file
    # to the target.
    if not os.path.isfile(scriptsdir + "/" + tfile):
        oldscript = fromdir + "/" + tfile
        print(oldscript)
        shutil.copy2(oldscript, scriptsdir)
    # If the target path exists compare the two scripts
    else:
        import difflib
        import esm_parser

        script_o = open(fromdir + "/" + tfile).readlines()
        script_t = open(scriptsdir + "/" + tfile).readlines()

        diffobj = difflib.SequenceMatcher(a=script_t, b=script_o)
        # If the files are different
        if not diffobj.ratio() == 1:
            # Find differences
            differences = (
                f"{fromdir + '/' + tfile} differs from "
                + f"{scriptsdir + '/' + tfile}:\n"
            )
            for line in color_diff(difflib.unified_diff(script_t, script_o)):
                differences += line

            # If the --update flag is used, notify that the target script will
            # be updated and do update it
            if gconfig["update"]:
                esm_parser.user_note(
                    f"Original {file_type} different from target",
                    differences + "\n" + f"{scriptsdir + '/' + tfile} will be updated!",
                )
                oldscript = fromdir + "/" + tfile
                print(oldscript)
                shutil.copy2(oldscript, scriptsdir)
            # If the --update flag is not called, exit with an error showing the
            # user how to proceed
            else:
                esm_parser.user_note(
                    f"Original {file_type} different from target",
                    differences
                    + "\n"
                    + "Note: You can choose to use -U flag in the esm_runscripts call "
                    + "to automatically update the runscript (WARNING: This "
                    + f"will overwrite your {file_type} in the experiment folder!)\n",
                )
                correct_input = False
                while not correct_input:
                    update_choice = input(
                        f"Do you want that {scriptsdir + '/' + tfile} is "
                        + "updated with the above changes? (y/n): "
                    )
                    if update_choice == "y":
                        correct_input = True
                        oldscript = fromdir + "/" + tfile
                        print(oldscript)
                        shutil.copy2(oldscript, scriptsdir)
                        print(f"{scriptsdir + '/' + tfile} updated!")
                    elif update_choice == "n":
                        correct_input = True
                        esm_parser.user_error(
                            f"Original {file_type} different from target",
                            differences
                            + "\n"
                            + "You can choose to -U flag in the esm_runscripts call "
                            + "to update the runscript without asking (WARNING: This "
                            + f"will overwrite your {file_type} in the experiment folder!)\n\n",
                        )
                    else:
                        print(f"'{update_choice}' is not a valid answer.")


def copy_tools_to_thisrun(config):
    """
    Copies the tools, namelists and runscripts to the experiment directory,
    making sure that they don't overwrite previously existing files unless
    the ``-U`` flag is used.
    Parameters
    ----------
    config : dict
        Dictionary containing the configuration information.
    """
    gconfig = config["general"]

    # Directory where the original runscript is
    fromdir = os.path.realpath(gconfig["started_from"])
    # Directory where the runscript is copied to
    scriptsdir = os.path.realpath(gconfig["experiment_scripts_dir"])
    # basedir + expid
    expdir = os.path.realpath(gconfig["experiment_dir"])

    # Paths inside the experiment directory where esm_tools and namelists
    # are copied to. Those are not functional but a reference to what was
    # the original state when the experiment was firstly started
    tools_dir = scriptsdir + "/esm_tools/functions"
    namelists_dir = scriptsdir + "/esm_tools/namelists"

    if config["general"]["verbose"]:
        print("Started from :", fromdir)
        print("Scripts Dir : ", scriptsdir)

    # Update namelists and esm_tools. These have no effect on the final
    # simulation as only the installed esm_tools with their runscripts
    # are used, unless esm_tools and namelists paths are manually defined
    # by the model or user (not recommended for the esm_tools path)
    if os.path.isdir(tools_dir) and gconfig["update"]:
        shutil.rmtree(tools_dir, ignore_errors=True)
    if os.path.isdir(namelists_dir) and gconfig["update"]:
        shutil.rmtree(namelists_dir, ignore_errors=True)

    # In case there is no esm_tools or namelists in the experiment folder,
    # copy from the default esm_tools path
    if not os.path.isdir(tools_dir):
        print("Copying standard yamls from: ", esm_rcfile.EsmToolsDir("FUNCTION_PATH"))
        esm_tools.copy_config_folder(tools_dir)
    if not os.path.isdir(namelists_dir):
        print(
            "Copying standard namelists from: ",
            esm_rcfile.EsmToolsDir("NAMELIST_PATH"),
        )
        esm_tools.copy_namelist_folder(namelists_dir)

    # check for recursive creation of the file tree. This prevents the risk of
    # creating a run directory tree inside the `scriptsdir`
    # example:
    # fromdir:    /mnt/lustre02/work/project/username/basedir
    # scriptsdir: /mnt/lustre02/work/project/username/basedir/expid/scripts
    # it is assumed that if 5th parent of the scriptsdir is inside the fromdir
    # then probably some sort of infinite recursion is entered. These lines
    # protect such problems
    scriptsdir_deep_parents = list(pathlib.Path(scriptsdir).parents)[5:]
    deep_nesting_found = pathlib.Path(expdir) in scriptsdir_deep_parents
    if deep_nesting_found:
        error_type = "runtime error"
        error_text = (
            f"deep recursion is detected in {__file__}:\n"
            f"- scriptsdir:         {scriptsdir}\n"
            f"- fromdir:            {fromdir}\n"
            f"- experiment dir:     {expdir}"
        )
        # exit right away to prevent further recursion. There might still be
        # running instances of esmr_runscripts and something like
        # `killall esm_runscripts` might be required
        esm_parser.user_error(error_type, error_text)

    # If ``fromdir`` and ``scriptsdir`` are the same, this is already a computing
    # simulation which means we want to use the script in the experiment folder,
    # so no copying is needed
    if (fromdir == scriptsdir) and not gconfig["update"]:
        if config["general"]["verbose"]:
            print("Started from the experiment folder, continuing...")
        return config
    # Not computing but initialisation
    else:
        if not fromdir == scriptsdir:
            if config["general"]["verbose"]:
                print("Not started from experiment folder, restarting...")
        else:
            print("Tools were updated, restarting...")

        # At this point, ``fromdir`` and ``scriptsdir`` are different. Update the
        # runscript if necessary
        update_runscript(
            fromdir, scriptsdir, gconfig["scriptname"], gconfig, "runscript"
        )

        # Update the ``additional_files`` if necessary
        for tfile in gconfig["additional_files"]:
            update_runscript(fromdir, scriptsdir, tfile, gconfig, "additional file")

        # remove the update option otherwise it will enter an infinite loop
        original_command = gconfig["original_command"]
        options_to_remove = [" -U ", " --update "]
        for option in options_to_remove:
            original_command = original_command.replace(option, " ")

        # Before resubmitting the esm_runscripts, the path of the runscript
        # needs to be modified. Remove the absolute/relative path
        runscript_absdir, runscript = os.path.split(gconfig["runscript_abspath"])
        original_command_list = original_command.split()
        new_command_list = []
        for command in original_command_list:
            # current command will contain the full path, so replace it with
            # the YAML file only since we are going to execute it from the
            # `scriptsdir` now
            if runscript in command:
                # gconfig['scriptname'] or `runscript` only contains the YAML file name
                command = runscript
            new_command_list.append(command)

        new_command = " ".join(new_command_list)
        restart_command = f"cd {scriptsdir}; esm_runscripts {new_command}"

        # prevent continuous addition of --no-motd
        if not "--no-motd" in restart_command:
            restart_command += " --no-motd "

        if config["general"]["verbose"]:
            print(restart_command)
        os.system(restart_command)

        gconfig["profile"] = False
        end_it_all(config)


def _copy_preliminary_files_from_experiment_to_thisrun(config):
    # I don't like this one bit. DB
    filelist = [
        (
            "scripts",
            f"{config['general']['expid']}_{config['general']['setup_name']}.date",
            "copy",
        )
    ]

    for filetype, filename, copy_or_link in filelist:
        source = config["general"]["experiment_" + filetype + "_dir"]
        dest = config["general"]["thisrun_" + filetype + "_dir"]
        if copy_or_link == "copy":
            method = shutil.copy2
        elif copy_or_link == "link":
            method = os.symlink
        if os.path.isfile(source + "/" + filename):
            method(source + "/" + filename, dest + "/" + filename)
    this_script = config["general"]["scriptname"]
    shutil.copy2("./" + this_script, config["general"]["thisrun_scripts_dir"])

    for additional_file in config["general"]["additional_files"]:
        shutil.copy2(additional_file, config["general"]["thisrun_scripts_dir"])
    return config


def _show_simulation_info(config):
    six.print_()
    six.print_(80 * "=")
    six.print_("STARTING SIMULATION JOB!")
    six.print_(f"Experiment ID = {config['general']['expid']}")
    six.print_(f"Setup = {config['general']['setup_name']}")
    if "coupled_setup" in config["general"]:
        six.print_("This setup consists of:")
        for model in config["general"]["valid_model_names"]:
            six.print_(f"- {model}")
    six.print_("Experiment is installed in:")
    six.print_(f"       {config['general']['base_dir']}/{config['general']['expid']}")
    six.print_(80 * "=")
    six.print_()
    return config



import os
import sys
import psutil
import time

from . import helpers
from . import database_actions
from . import logfiles


def run_job(config):
    helpers.evaluate(config, "observe", "observe_recipe")
    return config


def init_monitor_file(config):
    # which job am I spying on?
    observe_job = config["general"]["jobtype"]
    actual_job = observe_job.replace("observe_", "")
    called_from = actual_job
    # called_from = config["general"]["last_jobtype"]

    del config["general"]["command_line_config"]["launcher_pid"]

    exp_log_path = (
        config["general"]["experiment_log_dir"]
        + config["general"]["expid"]
        + "_"
        + config["general"]["setup_name"]
        + "_"
        + called_from
        + "_"
        + config["general"]["run_datestamp"]
        +
        # "_" +
        # str(config["general"]["jobid"]) +
        ".log"
    )
    log_in_run = (
        config["general"]["thisrun_log_dir"]
        + config["general"]["expid"]
        + "_"
        + config["general"]["setup_name"]
        + "_"
        + called_from
        +
        # "_" +
        # str(config["general"]["jobid"]) +
        ".log"
    )

    if os.path.isfile(exp_log_path):
        os.symlink(exp_log_path, log_in_run)

    monitor_file = logfiles.logfile_handle

    print(called_from)
    print(exp_log_path)

    monitor_file.write("observing job initialized \n")
    monitor_file.write(
        "attaching to process " + str(config["general"]["launcher_pid"]) + " \n"
    )
    monitor_file.write("Called from a " + called_from + "job \n")
    return config


def wait_and_observe(config):
    if config["general"]["submitted"]:
        monitor_file = logfiles.logfile_handle
        thistime = 0
        error_check_list = assemble_error_list(config)
        while job_is_still_running(config):
            monitor_file.write("still running \n")
            config["general"]["next_test_time"] = thistime
            config = check_for_errors(config)
            thistime = thistime + 10
            time.sleep(10)
        thistime = thistime + 100000000
        config["general"]["next_test_time"] = thistime
        config = check_for_errors(config)
    return config


def wake_up_call(config):
    monitor_file = logfiles.logfile_handle
    monitor_file.write("job ended, starting to tidy up now \n")
    return config


def assemble_error_list(config):
    gconfig = config["general"]

    thisjob = config["general"]["jobtype"]
    called_from = thisjob.replace("observe_", "")
    if not called_from == "compute":
        config["general"]["error_list"] = []
        return config

    known_methods = ["warn", "kill"]

    stdout = (
        gconfig["thisrun_log_dir"]
        + "/"
        + gconfig["expid"]
        + "_"
        + config["general"]["setup_name"]
        + "_"
        + called_from
        # + "_"
        # + gconfig["jobid"]
        + ".log"
    )

    error_list = [
        ("error", stdout, "warn", 60, 60, "keyword error detected, watch out")
    ]

    for model in config:
        if "check_error" in config[model]:
            for trigger in config[model]["check_error"]:
                search_file = stdout
                method = "warn"
                frequency = 60
                message = "keyword " + trigger + " detected, watch out"
                if isinstance(config[model]["check_error"][trigger], dict):
                    if "file" in config[model]["check_error"][trigger]:
                        search_file = config[model]["check_error"][trigger]["file"]
                        if search_file == "stdout" or search_file == "stderr":
                            search_file = stdout
                    if "method" in config[model]["check_error"][trigger]:
                        method = config[model]["check_error"][trigger]["method"]
                        if method not in known_methods:
                            method = "warn"
                    if "message" in config[model]["check_error"][trigger]:
                        message = config[model]["check_error"][trigger]["message"]
                    if "frequency" in config[model]["check_error"][trigger]:
                        frequency = config[model]["check_error"][trigger]["frequency"]
                        try:
                            frequency = int(frequency)
                        except:
                            frequency = 60
                elif isinstance(config[model]["check_error"][trigger], str):
                    pass
                else:
                    continue
                error_list.append(
                    (trigger, search_file, method, frequency, frequency, message)
                )
    config["general"]["error_list"] = error_list
    return config


def check_for_errors(config):
    thisjob = config["general"]["jobtype"]
    called_from = thisjob.replace("observe_", "")
    if not called_from == "compute":
        return config

    new_list = []
    error_check_list = config["general"]["error_list"]
    monitor_file = logfiles.logfile_handle
    time = config["general"]["next_test_time"]
    for (
        trigger,
        search_file,
        method,
        next_check,
        frequency,
        message,
    ) in error_check_list:
        warned = 0
        if next_check <= time:
            if os.path.isfile(search_file):
                with open(search_file) as origin_file:
                    for line in origin_file:
                        if trigger.upper() in line.upper():
                            if method == "warn":
                                warned = 1
                                monitor_file.write("WARNING: " + message + "\n")
                                break
                            elif method == "kill":
                                cancel_job = f"scancel {config['general']['jobid']}"
                                monitor_file.write("ERROR: " + message + "\n")
                                monitor_file.write("Will kill the run now..." + "\n")
                                monitor_file.flush()
                                print("ERROR: " + message)
                                print("Will kill the run now...", flush=True)
                                database_actions.database_entry_crashed(config)
                                os.system(cancel_job)
                                sys.exit(42)
            next_check += frequency
        if warned == 0:
            new_list.append(
                (trigger, search_file, method, next_check, frequency, message)
            )
    config["general"]["error_list"] = new_list
    return config


def job_is_still_running(config):
    if psutil.pid_exists(config["general"]["launcher_pid"]):
        return True
    return False



"""Top-level package for ESM Runscripts."""

__author__ = """Dirk Barbi"""
__email__ = "dirk.barbi@awi.de"
__version__ = "6.0.3"

from .sim_objects import *
from .batch_system import *
from .database import *
from .database_actions import *
from .prepcompute import *
from .tidy import *
from .prepare import *
from .last_minute import *
from .dataprocess import *
from .filelists import *
from .chunky_parts import *
from .workflow import *
from .tidy import *
from .namelists import *
from .virtual_env_builder import *
from .prepexp import *
